/**************************************************************/
/*                                                            */
/*  vge.c                                                     */
/*  Voxel Geometry Engine                                     */
/*  (c)1998 Marek Uher, xuher@sun.felk.cvut.cz                */
/*  http://service.felk.cvut.cz/~xuher/                       */
/*  Czech Technical University (CTU), Prague, Czech Republic  */
/*                                                            */
/*  vge v0.86 (November 1998)                                 */
/*                                                            */
/**************************************************************/


/*                                                            */
/*  Standard libraries include files                          */
/*                                                            */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <GL/glut.h>


/*                                                            */
/*  vge include file                                          */
/*                                                            */
#include "vge.h"


/*                                                            */
/*  Debug file                                                */
/*                                                            */
#define DEBUG_FILE "debug.map"


/*                                                            */
/*  line size (needed for tokens from texture file)           */
/*                                                            */
#define LINE_SIZE 25

/*                                                            */
/*  Freq. table                                               */
/*                                                            */
#define NaC 300
int freqTable[8][2] = { { NaC, 0 }, { NaC, 0 }, { NaC, 0 }, { NaC, 0 },
			{ NaC, 0 }, { NaC, 0 }, { NaC, 0 }, { NaC, 0 } };


/*                                                            */
/*  Voxel objects in memory                                   */
/*                                                            */
VoxelObject *voxelObject;
VoxelObject *voxelObjectLod1;
VoxelObject *voxelObjectLod2;
VoxelObject *voxelObjectLod3;
VoxelObject *voxelObjectLod4;


/*                                                            */
/*  Texture file and debug file                               */
/*                                                            */
FILE *textureFile;
FILE *faceFile;


/*                                                            */
/*  Variables for parsing texture file                        */
/*                                                            */
char tmpChar;
char line[LINE_SIZE + 1];
char token[4];


/*                                                            */
/*  Time variables for profiling                              */
/*                                                            */
time_t startTime;
time_t endTime;
time_t *timeStartInterval;
time_t *timeEndInterval;


/*                                                            */
/*  Texture file version and release                          */
/*                                                            */
Version version;
Release release;


/*                                                            */
/*  Voxel object resolution (bounding box)                    */
/*                                                            */
Resolution x, y, z;


/*                                                            */
/*  Level of detail                                           */
/*                                                            */
LOD lod = 0;


/*                                                            */
/*  Flag - (non)compressed data stream                        */
/*                                                            */
CDSflag cds = NON_COMPRESED_DATA_STREAM;


/*                                                            */
/*  Color palette and color index                             */
/*                                                            */
ColorPalette colorPalette[PALETTE_SIZE];
PaletteIndex palIndex = 1;
Color newColor;


/*                                                            */
/*  Voxel buffer (for temp data)                              */
/*                                                            */
VoxelBuffer voxelBuffer;
Voxel voxel = EMPTY_VOXEL;


/*                                                            */
/*  Surface (quad)                                            */
/*                                                            */
Quad qquad;


/*                                                            */
/*  Statistic variables                                       */
/*                                                            */
Counter maxNumOfInputQuads = 0;
Counter realNumOfInputQuads = 0;
Counter realNumOfOutputQuads = 0;
Ratio ratio = 0.0;


/*                                                            */
/*  Surface normals                                           */
/*                                                            */
/*  Step: tranparent -> solid                                 */
/*                                                            */
const GLfloat normal_0C_x[3] = { -1.0,  0.0,  0.0 };
const GLfloat normal_0C_y[3] = {  0.0, -1.0,  0.0 };
const GLfloat normal_0C_z[3] = {  0.0,  0.0, -1.0 };


/*                                                            */
/* Step: solid -> transparent                                 */
/*                                                            */
const GLfloat normal_C0_x[3] = { 1.0, 0.0, 0.0 };
const GLfloat normal_C0_y[3] = { 0.0, 1.0, 0.0 };
const GLfloat normal_C0_z[3] = { 0.0, 0.0, 1.0 };


/*                                                            */
/*  New resolutions of voxel object (need for LOD)            */
/*                                                            */
Resolution newX;
Resolution newY;
Resolution newZ;

GLfloat _amb[]		= { 0.0,   0.0,   0.0,   0.0 };
GLfloat _use[]		= { 1.0,   1.0,   1.0,   1.0 };
GLfloat _specular[]	= { 0.394, 0.272, 0.167, 1.0 };
GLfloat _shininess[]	= { 20 };

/*                                                            */
/*  Write quad to file (debuging, normaly disabled)           */
/*                                                            */
void WriteQuad(Resolution x1, Resolution y1, Resolution z1, Resolution x2, Resolution y2, Resolution z2, Color color, const GLfloat *normal){

	/*fprintf(faceFile, "%d %d %d %d %d %d %d\n", x1, y1, z1, x2, y2, z2, color);*/
    _use[0] = voxelObject->colorPalette[color].red/255.0;
    _use[1] = voxelObject->colorPalette[color].green/255.0;
    _use[2] = voxelObject->colorPalette[color].blue/255.0;
	
    glMaterialfv(GL_FRONT, GL_AMBIENT, _amb);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, _use); 
    glMaterialfv(GL_FRONT, GL_SPECULAR, _specular); 
    glMaterialfv(GL_FRONT, GL_SHININESS, _shininess); 
	if (z1==z2){
		if (normal[2] == 1){
		glBegin(GL_QUADS);
		  glNormal3f(0.0, 0.0, 1.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x2, y1, z1);
		  glVertex3f(x2, y2, z1);
		  glVertex3f(x1, y2, z1);
		glEnd();
		}
		else{
		glBegin(GL_QUADS);
		  glNormal3f(0.0, 0.0, -1.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x1, y2, z1);
		  glVertex3f(x2, y2, z1);
		  glVertex3f(x2, y1, z1);
		glEnd();
		}
	}

	if (y1==y2){
		if (normal[1] == -1){
		glBegin(GL_QUADS);
		  glNormal3f(0.0, 1.0, 0.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x2, y1, z1);
		  glVertex3f(x2, y1, z2);
		  glVertex3f(x1, y1, z2);
		glEnd();
		}
		else{
		glBegin(GL_QUADS);
		  glNormal3f(0.0, -1.0, 0.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x1, y1, z2);
		  glVertex3f(x2, y1, z2);
		  glVertex3f(x2, y1, z1);
		glEnd();
		}
	}

	if (x1==x2){
		if (normal[0] == -1){
		glBegin(GL_QUADS);
		  glNormal3f(1.0, 0.0, 0.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x1, y1, z2);
		  glVertex3f(x1, y2, z2);
		  glVertex3f(x1, y2, z1);
		glEnd();
		}
		else{
		glBegin(GL_QUADS);
		  glNormal3f(-1.0, 0.0, 0.0);
		  glVertex3f(x1, y1, z1);
		  glVertex3f(x1, y2, z1);
		  glVertex3f(x1, y2, z2);
		  glVertex3f(x1, y1, z2);
		glEnd();
		}
	}
	return;
}


/*                                                            */
/*  Test if x is even or odd                                  */
/*                                                            */
int odd(int x){

	return(((((int)((float)x / 2.0)) * 2) == x) ? 0 : 1);
}


/*                                                            */
/*  Function which initialize voxel slice to value            */
/*                                                            */
void InitVoxelSlice(VoxelSlice slice, Resolution x, Resolution y, Value value){

	register Counter i, j;
	

	for(i = 0; i < x; i++){
		for(j = 0; j < y; j++){
			slice[i][j] = value;
		}
	}
	return;
}


/*                                                            */
/*  Function which initialize voxel buffer to value           */
/*                                                            */
void InitVoxelBuffer(VoxelBuffer voxelBuffer, Resolution x, Resolution y, Resolution z, Value value){

	register Counter i, j, k;
	

	for(i = 0; i < x; i++){
		for(j = 0; j < y; j++){
			for(k = 0; k < z; k++){
				voxelBuffer[i][j][k] = value;
			}
		}
	}
	return;
}


/*                                                            */
/*  Function for creating voxel slice                         */
/*                                                            */  
VoxelSlice CreateVoxelSlice(Resolution x, Resolution y){

	VoxelSlice newVoxelSlice;
	register Index i = 0;
	register Index j = 0;


	if ((newVoxelSlice = (Voxel **) malloc(x * sizeof(Voxel *))) == NULL){
		CatchException;
	}

	for(i = 0; i < x; i++){
		if ((newVoxelSlice[i] = (Voxel *) malloc(y * sizeof(Voxel))) == NULL){
			CatchException;
		}
	}

	InitVoxelSlice(newVoxelSlice, x, y, EMPTY_VOXEL);

	return(newVoxelSlice);
}


/*                                                            */
/*  Function for creating voxel buffer                        */
/*                                                            */
VoxelBuffer CreateVoxelBuffer(Resolution x, Resolution y, Resolution z){

	VoxelBuffer newVoxelBuffer;
	register Index i = 0;
	register Index j = 0;
	register Index k = 0;


	if ((newVoxelBuffer = (Voxel ***) malloc(x * sizeof(Voxel **))) == NULL){
		CatchException;
	}

	for(i = 0; i < x; i++){
		if ((newVoxelBuffer[i] = (Voxel **) malloc(y * sizeof(Voxel *))) == NULL){
			CatchException;
		}
	}

	for(i = 0; i < x; i++){
		for(j = 0; j < y; j++){
			if ((newVoxelBuffer[i][j] = (Voxel *) malloc(z * sizeof(Voxel))) == NULL){
				CatchException;
			}
		}
	}

	InitVoxelBuffer(newVoxelBuffer, x, y, z, EMPTY_VOXEL);

	return(newVoxelBuffer);
}


/*                                                            */
/*  Function for creating voxel object in memory              */
/*                                                            */
VoxelObject *CreateVoxelObject(char *name, Version ver, Release rel, Voxel x, Voxel y, Voxel z, LOD lod, CDSflag cdsFlag, ColorPalette colorPalette[]){

	VoxelBuffer tempBuffer;
	VoxelObject *newObject;
	register Index i;


	if ((newObject = (VoxelObject *) malloc(sizeof(VoxelObject))) == NULL){
		CatchException;
	}

	tempBuffer = CreateVoxelBuffer(x, y, z);
	strcpy(newObject -> objectName, name);
	newObject -> version = ver;
	newObject -> release = rel;
	newObject -> resX = x;
	newObject -> resY = y;
	newObject -> resZ = z;
	newObject -> lod = lod;
	newObject -> cdsFlag = cdsFlag;
	for(i = 0; i < PALETTE_SIZE; i++){
		newObject -> colorPalette[i].red = colorPalette[i].red;
		newObject -> colorPalette[i].green = colorPalette[i].green;
		newObject -> colorPalette[i].blue = colorPalette[i].blue;
	}
	newObject -> voxelBuffer = tempBuffer;

	return(newObject);
}


/*                                                            */
/*  Release voxel object                                      */
/*                                                            */
void ReleaseVoxelObject(VoxelObject *voxObject){

	char *noname = "Free memory";
	register Index i;


	strcpy(voxObject -> objectName, noname);
	voxObject -> version = 0;
	voxObject -> release = 0;
	voxObject -> resX = 0;
	voxObject -> resY = 0;
	voxObject -> resZ = 0;
	voxObject -> lod = 0;
	voxObject -> cdsFlag = 0;
	for(i = 0; i < PALETTE_SIZE; i++){
		voxObject -> colorPalette[i].red = colorPalette[i].red;
		voxObject -> colorPalette[i].green = colorPalette[i].green;
		voxObject -> colorPalette[i].blue = colorPalette[i].blue;
	}
	ReleaseVoxelBuffer(voxObject -> voxelBuffer);
	free((void *) voxObject);
	return;
}


/*                                                            */
/*  Parse texture file for tokens and data                    */
/*                                                            */
void ParseTextureFile(void){

	register Index i, j, k;


	while (!feof(textureFile)){
	
		tmpChar = fgetc(textureFile);

		switch(tmpChar){
		
			case '/': 
			
				do {
					tmpChar = fgetc(textureFile);
				} while (tmpChar != '/');
				tmpChar = fgetc(textureFile);
				ParseTextureFile();
				
				break;

			case '#':
			
				fgets(line, LINE_SIZE, textureFile);
				sscanf(line, "%3s", &token);

				if (!strcmp(token, "vtf")) {
					printf("%s\n", "Voxel texture format");
					printf("%s\n", "Reading voxel data sream, please, stand by ...");
					fflush(stdout);
				}

				if (!strcmp(token, "ver")) {
					sscanf(line, "%s %d.%d", &token, &version, &release);

				}

				if (!strcmp(token, "res")) {
					sscanf(line, "%s %d %d %d", &token, &x, &y, &z);
				}
	
				if (!strcmp(token, "lod")) {
					sscanf(line, "%s %d", &token, &lod);
					
				}

				if (!strcmp(token, "cds")) {
					sscanf(line, "%s %d", &token, &cds);
				}
				
				if (!strcmp(token, "vds")) {
					voxel = fgetc(textureFile);
					if (voxel == '/') {
						do {
							voxel = fgetc(textureFile);
						} while (voxel != '/');
						voxel = fgetc(textureFile);
					}
					else {
						ungetc(voxel, textureFile);
					}

					voxelBuffer = CreateVoxelBuffer(x, y, z);
					for(k = 0; k < z; k++){
						for(i = 0; i < x; i++){
							for(j = 0; j < y; j++){
								fscanf(textureFile, "%d", &voxel);
								voxelBuffer[i][j][k] = voxel;
							}
						} 
					}		
				}
				
				ParseTextureFile();
				
				break;

			case 'c':
			
				if (palIndex < PALETTE_SIZE) {
					fscanf(textureFile, "%d %d %d", &colorPalette[palIndex].red, &colorPalette[palIndex].green, &colorPalette[palIndex].blue);
					palIndex++;
				}
				ParseTextureFile();
				
				break;
				
			default:
			
				break;
		}
	}
	return;
}


/*                                                            */
/*  Find last clean index in voxel map                        */
/*                                                            */
void LastClean(Resolution *x, Resolution *y, Resolution maxI, Resolution maxJ, VoxelSlice map){

	register Index i, j;
	

	for(i = 0; i < maxI; i++){
		for(j = 0; j < maxJ; j++){
			if (map[i][j] == CLEAN){
				*x = i;
				*y = j;
				return;
			}
		}
	}
}


/*                                                            */
/*  Test if exist some clean position on voxel map            */
/*                                                            */
int IsClean(Resolution maxI, Resolution maxJ, VoxelSlice testedSlice){

	register Index i, j;
	

	for (i = 0; i < maxI; i++){
		for (j = 0; j < maxJ; j++){
			if (testedSlice[i][j] == CLEAN){
				return(DIRTY);
			}
		}
	}
	return(CLEAN);	
}


/*                                                            */
/*  Mark index as dirty on index x,y in voxel map             */
/*                                                            */
void MarkDirty(Resolution x, Resolution y, VoxelSlice map){

	map[x][y] = DIRTY;
	return;
}	


/*                                                            */
/*  Find solid quads on voxel slice                           */
/*                                                            */ 
void FindQuad(Resolution x, Resolution y, Resolution maxX, Resolution maxY, VoxelSlice slice, VoxelSlice dirtyMap, Quad *qquad){

	Resolution quadHeight = 0;
	Resolution quadWidth = 0;
	Resolution x1, y1, x2, y2;
	Color color;
	register int i, j;
	Value flagDirection = X_AND_Y;
	Direction direction = RIGHT;
	Value growable;


	/*  start generate growable area  */
	quadWidth = quadHeight = 1;

	/*  first corner of quad (constant)  */
	x1 = x;
	y1 = y;

	/*  color of quad  */
	color = slice[x1][y1];

	dirtyMap[x1][y1] = DIRTY;

	while (flagDirection > ANY) {

		x2 = x1 + quadWidth;
		y2 = y1 + quadHeight;

		growable = NO;

		if (direction == RIGHT) {

			if ((x1 + quadWidth) >= maxX) {
				growable = YES;
			}

			else {
				for(j = y1; j < (y1 + quadHeight); j++) {
					if(slice[x1 + quadWidth][j] != color || dirtyMap[x1 + quadWidth][j] == DIRTY) {
						growable = YES;
					}
				}
			}
		}

		else {

			if ((y1 + quadHeight) >= maxY) {
				growable = YES;
			}

			else {
				for(i = x1; i < (x1 + quadWidth); i++) {
					if (slice[i][y1 + quadHeight] != color || dirtyMap[i][y1 + quadHeight] == DIRTY) {
						growable = YES;
					}
				}
			}
		}

		/* growable is 0 when may go to selected direction */
		if (growable == ANY) {
			if (direction == RIGHT) {
				quadWidth++;
				for(j = y1; j < (y1 + quadHeight); j++) {
					MarkDirty(x1 + quadWidth - 1, j, dirtyMap);
				}
			}
			else {
				quadHeight++;
				for(i = x1; i < x1 + quadWidth; i++) {
					MarkDirty(i, y1 + quadHeight - 1, dirtyMap);
				}
			}
		}

		if (flagDirection == X_AND_Y) {
			direction = (direction == RIGHT) ? DOWN : RIGHT;
		}

		if (growable != ANY) {
			flagDirection--;
		}
	}

	qquad -> x1 = x1;
	qquad -> y1 = y1;
	qquad -> x2 = x1 + quadWidth;
	qquad -> y2 = y1 + quadHeight;
	qquad -> color = color;
	realNumOfOutputQuads++;
	return;
}


/*                                                            */
/*  Convert voxel to solid faces                              */
/*                                                            */
void Voxel2Faces(Direction direction){

	VoxelSlice previousSlice;
	VoxelSlice currentSlice;
	VoxelSlice dirtyMap_0C;
	VoxelSlice dirtyMap_C0;
	VoxelSlice quadSlice_0C;
	VoxelSlice quadSlice_C0;
		

	register Index i, j, k;
	Resolution x, y;
	Resolution x1, y1, z1;
	Resolution x2, y2, z2;
	Color color;
	
	switch(direction){
	
		case DIRECTION_X:
	
			previousSlice = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ);
			currentSlice  = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ);
			dirtyMap_0C   = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ);
			dirtyMap_C0   = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ);
			quadSlice_0C  = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ);
			quadSlice_C0  = CreateVoxelSlice(voxelObject -> resY, voxelObject -> resZ); 
			
			for(i = 0; i <= voxelObject -> resX; i++){

				InitVoxelSlice(dirtyMap_0C,  voxelObject -> resY, voxelObject -> resZ, CLEAN);
				InitVoxelSlice(dirtyMap_C0,  voxelObject -> resY, voxelObject -> resZ, CLEAN);
				InitVoxelSlice(quadSlice_0C, voxelObject -> resY, voxelObject -> resZ, EMPTY_VOXEL);
				InitVoxelSlice(quadSlice_C0, voxelObject -> resY, voxelObject -> resZ, EMPTY_VOXEL);
								
				if (i == 0) {
					for(j = 0; j < voxelObject -> resY; j++){
						for(k = 0; k < voxelObject -> resZ; k++){
							previousSlice[j][k] = EMPTY_VOXEL;
							currentSlice[j][k] = voxelObject -> voxelBuffer[i][j][k];
						}
					}
				}
				else {		
					if (i == voxelObject -> resX){
						for(j = 0; j < voxelObject -> resY; j++){
							for(k = 0; k < voxelObject -> resZ; k++){
								previousSlice[j][k] = voxelObject -> voxelBuffer[i-1][j][k];
								currentSlice[j][k] = EMPTY_VOXEL;
							} 
						}
					}
					else {
						for(j = 0; j < voxelObject -> resY; j++){
							for(k = 0; k < voxelObject -> resZ; k++){
								previousSlice[j][k] = voxelObject -> voxelBuffer[i-1][j][k];
								currentSlice[j][k] = voxelObject -> voxelBuffer[i][j][k];
							}
						}
					}	
				}				
				
				for(j = 0; j < voxelObject -> resY; j++){
					for(k = 0; k < voxelObject -> resZ; k++){

						/*  transparent voxel -> transparent voxel  or  */
						/*  color (solid) voxel -> color (solid) voxel  */
						if (((previousSlice[j][k] == EMPTY_VOXEL) && (currentSlice[j][k] == EMPTY_VOXEL)) 
						 || ((previousSlice[j][k] != EMPTY_VOXEL) && (currentSlice[j][k] != EMPTY_VOXEL))){
							quadSlice_0C[j][k] = EMPTY_VOXEL;
							quadSlice_C0[j][k] = EMPTY_VOXEL;
							dirtyMap_0C[j][k]  = DIRTY;
							dirtyMap_C0[j][k]  = DIRTY;
						}
						
						/*  transparent voxel -> color (solid) voxel  */ 
						if ((previousSlice[j][k] == EMPTY_VOXEL) && (currentSlice[j][k] != EMPTY_VOXEL)){
							quadSlice_0C[j][k] = currentSlice[j][k];
							quadSlice_C0[j][k] = EMPTY_VOXEL;
							dirtyMap_C0[j][k]  = DIRTY;
						}
						
						/*  color (solid) voxel -> transparent voxel  */
						if ((previousSlice[j][k] != EMPTY_VOXEL) && (currentSlice[j][k] == EMPTY_VOXEL)){
							quadSlice_0C[j][k] = EMPTY_VOXEL;
							quadSlice_C0[j][k] = previousSlice[j][k];
							dirtyMap_0C[j][k]  = DIRTY;
						}
					}
				}
				/*glNormal3fv(normal_0C_x);*/
				while (IsClean(voxelObject -> resY, voxelObject -> resZ, dirtyMap_0C)){
					LastClean(&x, &y, voxelObject -> resY, voxelObject -> resZ, dirtyMap_0C);
					FindQuad(x, y, voxelObject -> resY, voxelObject -> resZ, quadSlice_0C, dirtyMap_0C, &qquad);
					/* printf("%i %i\n", x, y); */
					x1 = x2 = i;
					y1 = qquad.x1;
					z1 = qquad.y1;
					y2 = qquad.x2;
					z2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_0C_x);
				}

				/*glNormal3fv(normal_C0_x);*/
				while (IsClean(voxelObject -> resY, voxelObject -> resZ, dirtyMap_C0)){
					LastClean(&x, &y, voxelObject -> resY, voxelObject -> resZ, dirtyMap_C0);
					FindQuad(x, y, voxelObject -> resY, voxelObject -> resZ, quadSlice_C0, dirtyMap_C0, &qquad);
					x1 = x2 = i;
					y1 = qquad.x1;
					z1 = qquad.y1;
					y2 = qquad.x2;
					z2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_C0_x);
				}
			}
									
			break;

		case DIRECTION_Y:

			previousSlice = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ);
			currentSlice  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ);
			dirtyMap_0C   = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ);
			dirtyMap_C0   = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ);
			quadSlice_0C  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ);
			quadSlice_C0  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resZ); 

			for(j = 0; j <= voxelObject -> resY; j++){

				InitVoxelSlice(dirtyMap_0C,  voxelObject -> resX, voxelObject -> resZ, CLEAN);
				InitVoxelSlice(dirtyMap_C0,  voxelObject -> resX, voxelObject -> resZ, CLEAN);
				InitVoxelSlice(quadSlice_0C, voxelObject -> resX, voxelObject -> resZ, EMPTY_VOXEL);
				InitVoxelSlice(quadSlice_C0, voxelObject -> resX, voxelObject -> resZ, EMPTY_VOXEL);

				if (j == 0) {
					for(i = 0; i < voxelObject -> resX; i++){
						for(k = 0; k < voxelObject -> resZ; k++){
							previousSlice[i][k] = EMPTY_VOXEL;
							currentSlice[i][k] = voxelObject -> voxelBuffer[i][j][k];
						}
					}
				}
				else {		
					if (j == voxelObject -> resY){
						for(i = 0; i < voxelObject -> resX; i++){
							for(k = 0; k < voxelObject -> resZ; k++){
								previousSlice[i][k] = voxelObject -> voxelBuffer[i][j-1][k];
								currentSlice[i][k] = EMPTY_VOXEL;
							}
						}
					}
					else {
						for(i = 0; i < voxelObject -> resX; i++){
							for(k = 0; k < voxelObject -> resZ; k++){
								previousSlice[i][k] = voxelObject -> voxelBuffer[i][j-1][k];
								currentSlice[i][k] = voxelObject -> voxelBuffer[i][j][k];
							}
						}
					}	
				}				
				
				for(i = 0; i < voxelObject -> resX; i++){
					for(k = 0; k < voxelObject -> resZ; k++){

						/*  transparent voxel -> transparent voxel  or  */
						/*  color (solid) voxel -> color (solid) voxel  */
						if (((previousSlice[i][k] == EMPTY_VOXEL) && (currentSlice[i][k] == EMPTY_VOXEL)) 
						 || ((previousSlice[i][k] != EMPTY_VOXEL) && (currentSlice[i][k] != EMPTY_VOXEL))){
							quadSlice_0C[i][k] = EMPTY_VOXEL;
							quadSlice_C0[i][k] = EMPTY_VOXEL;
							dirtyMap_0C[i][k]  = DIRTY;
							dirtyMap_C0[i][k]  = DIRTY;
						}
						
						/*  transparent voxel -> color (solid) voxel  */ 
						if ((previousSlice[i][k] == EMPTY_VOXEL) && (currentSlice[i][k] != EMPTY_VOXEL)){
							quadSlice_0C[i][k] = currentSlice[i][k];
							quadSlice_C0[i][k] = EMPTY_VOXEL;
							dirtyMap_C0[i][k]  = DIRTY;
						}
						
						/*  color (solid) voxel -> transparent voxel  */
						if ((previousSlice[i][k] != EMPTY_VOXEL) && (currentSlice[i][k] == EMPTY_VOXEL)){
							quadSlice_0C[i][k] = EMPTY_VOXEL;
							quadSlice_C0[i][k] = previousSlice[i][k];
							dirtyMap_0C[i][k]  = DIRTY;
						}
					}
				}
				
				/*glNormal3fv(normal_0C_y);*/
				while (IsClean(voxelObject -> resX, voxelObject -> resZ, dirtyMap_0C)){
					LastClean(&x, &y, voxelObject -> resX, voxelObject -> resZ, dirtyMap_0C);
					FindQuad(x, y, voxelObject -> resX, voxelObject -> resZ, quadSlice_0C, dirtyMap_0C, &qquad);
					y1 = y2 = j;
					x1 = qquad.x1;
					z1 = qquad.y1;
					x2 = qquad.x2;
					z2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_0C_y);
				}

				/*glNormal3fv(normal_C0_y);*/
				while (IsClean(voxelObject -> resX, voxelObject -> resZ, dirtyMap_C0)){
					LastClean(&x, &y, voxelObject -> resX, voxelObject -> resZ, dirtyMap_C0);
					FindQuad(x, y, voxelObject -> resX, voxelObject -> resZ, quadSlice_C0, dirtyMap_C0, &qquad);
					y1 = y2 = j;
					x1 = qquad.x1;
					z1 = qquad.y1;
					x2 = qquad.x2;
					z2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_C0_y);
				}
			}

			break;

		case DIRECTION_Z:

			previousSlice = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY);
			currentSlice  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY);
			dirtyMap_0C   = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY);
			dirtyMap_C0   = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY);
			quadSlice_0C  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY);
			quadSlice_C0  = CreateVoxelSlice(voxelObject -> resX, voxelObject -> resY); 

			for(k = 0; k <= voxelObject -> resZ; k++){
				
				InitVoxelSlice(dirtyMap_0C,  voxelObject -> resX, voxelObject -> resY, CLEAN);
				InitVoxelSlice(dirtyMap_C0,  voxelObject -> resX, voxelObject -> resY, CLEAN);
				InitVoxelSlice(quadSlice_0C, voxelObject -> resX, voxelObject -> resY, EMPTY_VOXEL);
				InitVoxelSlice(quadSlice_C0, voxelObject -> resX, voxelObject -> resY, EMPTY_VOXEL);

				if (k == 0) {
					for(i = 0; i < voxelObject -> resX; i++){
						for(j = 0; j < voxelObject -> resY; j++){
							previousSlice[i][j] = EMPTY_VOXEL;
							currentSlice[i][j] = voxelObject -> voxelBuffer[i][j][k];
						}
					}
				}
				else {		
					if (k == voxelObject -> resZ){
						for(i = 0; i < voxelObject -> resX; i++){
							for(j = 0; j < voxelObject -> resY; j++){
								previousSlice[i][j] = voxelObject -> voxelBuffer[i][j][k-1];
								currentSlice[i][j] = EMPTY_VOXEL;
							}
						}
					}
					else {
						for(i = 0; i < voxelObject -> resX; i++){
							for(j = 0; j < voxelObject -> resY; j++){
								previousSlice[i][j] = voxelObject -> voxelBuffer[i][j][k-1];
								currentSlice[i][j] = voxelObject -> voxelBuffer[i][j][k];
							}
						}
					}	
				}				
				
				for(i = 0; i < voxelObject -> resX; i++){
					for(j = 0; j < voxelObject -> resY; j++){

						/*  transparent voxel -> transparent voxel  or  */
						/*  color (solid) voxel -> color (solid) voxel  */
						if (((previousSlice[i][j] == EMPTY_VOXEL) && (currentSlice[i][j] == EMPTY_VOXEL)) 
						 || ((previousSlice[i][j] != EMPTY_VOXEL) && (currentSlice[i][j] != EMPTY_VOXEL))){
							quadSlice_0C[i][j] = EMPTY_VOXEL;
							quadSlice_C0[i][j] = EMPTY_VOXEL;
							dirtyMap_0C[i][j]  = DIRTY;
							dirtyMap_C0[i][j]  = DIRTY;
						}
						
						/*  transparent voxel -> color (solid) voxel  */ 
						if ((previousSlice[i][j] == EMPTY_VOXEL) && (currentSlice[i][j] != EMPTY_VOXEL)){
							quadSlice_0C[i][j] = currentSlice[i][j];
							quadSlice_C0[i][j] = EMPTY_VOXEL;
							dirtyMap_C0[i][j]  = DIRTY;
						}
						
						/*  color (solid) voxel -> transparent voxel  */
						if ((previousSlice[i][j] != EMPTY_VOXEL) && (currentSlice[i][j] == EMPTY_VOXEL)){
							quadSlice_0C[i][j] = EMPTY_VOXEL;
							quadSlice_C0[i][j] = previousSlice[i][j];
							dirtyMap_0C[i][j]  = DIRTY;
						}
					}
				}
				
				/*glNormal3fv(normal_0C_z);*/
				while (IsClean(voxelObject -> resX, voxelObject -> resY, dirtyMap_0C)){
					LastClean(&x, &y, voxelObject -> resX, voxelObject -> resY, dirtyMap_0C);
					FindQuad(x, y, voxelObject -> resX, voxelObject -> resY, quadSlice_0C, dirtyMap_0C, &qquad);
					z1 = z2 = k;
					x1 = qquad.x1;
					y1 = qquad.y1;
					x2 = qquad.x2;
					y2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_0C_z);
				}

				/*glNormal3fv(normal_C0_z);*/
				while (IsClean(voxelObject -> resX, voxelObject -> resY, dirtyMap_C0)){
					LastClean(&x, &y, voxelObject -> resX, voxelObject -> resY, dirtyMap_C0);
					FindQuad(x, y, voxelObject -> resX, voxelObject -> resY, quadSlice_C0, dirtyMap_C0, &qquad);
					z1 = z2 = k;
					x1 = qquad.x1;
					y1 = qquad.y1;
					x2 = qquad.x2;
					y2 = qquad.y2; 
					color = qquad.color;
					
					WriteQuad(x1, y1, z1, x2, y2, z2, color, normal_C0_z);
				}
			}
			
			break;
			
		default:
			fprintf(stderr, "%s\n", "wrong sliceing direction");  
			
			break;
	}

	ReleaseVoxelSlice(previousSlice);
	ReleaseVoxelSlice(currentSlice);
	ReleaseVoxelSlice(dirtyMap_0C);
	ReleaseVoxelSlice(dirtyMap_C0);
	ReleaseVoxelSlice(quadSlice_0C);
	ReleaseVoxelSlice(quadSlice_C0);
	return;
}

/*                                                            */
/*  Sorting function                                          */
/*                                                            */
int SortFunction( const void *x, const void *y){

	return(((*((int *)(x))) < (*((int *)(y)))) ? -1 : 1);
}

/*                                                            */
/*  Compute level of detail (LOD)                             */
/*                                                            */
void ComputeLOD(LOD nLod, char *objName){
/*                                                            */
/*  freqTable's member handlers                               */
/*                                                            */
#ifdef COLOR
	#define OLD_COLOR COLOR
	#undef  COLOR
	#define COLOR  0
#else
	#define COLOR  0
#endif

#ifdef FREQ
	#define OLD_FREQ FREQ
	#undef  FREQ
	#define FREQ   1
#else
	#define FREQ   1
#endif
	
	/*                                                    */
	/*  !!! IMPOTRTANT !!!                                */
	/*  The value of LOD_WEIGHT_VAL control the           */
	/*  computation of level of detail (LOD)              */
	/*  WEIGHT determine the influence of EMPTY_VOXEL     */
	/*                                                    */
#define LOD_WEIGHT_VAL 5 /* value is between 0 - 8            */
			 /* (value has effect only for 1 - 7) */   
#ifdef WEIGHT
	#define OLD_WEIGHT WEIGHT
	#undef  WEIGHT
	#define WEIGHT LOD_WEIGHT_VAL
#else
	#define WEIGHT LOD_WEIGHT_VAL
#endif
	
	register Counter i, j, k;
	register Counter ii, jj, kk;
	register Index   ni, nj, nk;
	int colorValue[8] = { NaC, NaC, NaC, NaC, NaC, NaC, NaC, NaC };
	register Index index = 0;
	register Index tableIndex = 0;
	float averageColor = 0.0;
	int freq = 0;
	struct _maxFreqColor_ {
		int color;
		int freq;
	} maxFreqColor;


	switch (nLod){
	
		case LOD_1: 
		
			printf("%s%d%s\n", "Computing level of detail (level ", nLod, ") ...");
			if (odd(voxelObject -> resX)) {
				newX = ((voxelObject -> resX) + 1) >> 1;
			}
			else {
				newX = (voxelObject -> resX) >> 1;
			}
			
			if (odd(voxelObject -> resY)) {
				newY = ((voxelObject -> resY) + 1) >> 1;
			}
			else {
				newY = (voxelObject -> resY) >> 1;
			}
			
			if (odd(voxelObject -> resZ)) {
				newZ = ((voxelObject -> resZ) + 1) >> 1;
			}
			else {
				newZ = (voxelObject -> resZ) >> 1;
			}
			
			voxelObjectLod1 = CreateVoxelObject(objName, version, release,  newX, newY, newZ, LOD_1, cds, colorPalette);

			printf("New resolution:\tx = %d\ty = %d\tz = %d\n", voxelObjectLod1 -> resX, voxelObjectLod1 -> resY, voxelObjectLod1 -> resZ);
			
			for(k = 0; k < voxelObjectLod1 -> resZ; k++){
			  for(i = 0; i < voxelObjectLod1 -> resX; i++){
			    for(j = 0; j < voxelObjectLod1 -> resY; j++){
			      
			      for (index = 0; index < 8; index++) {
				colorValue[index] = NaC;
				freqTable[index][COLOR] = NaC;
				freqTable[index][FREQ]  = 0;
			      }						
					  
						newColor = EMPTY_VOXEL;
						index = 0;
						tableIndex = 0;

						for (kk = 2*k; kk < (2*k + 2); kk++){
							for (ii = 2*i; ii < (2*i + 2); ii++){
								for (jj = 2*j; jj < (2*j + 2); jj++){

									if (ii < voxelObject -> resX) {
										ni = ii;
									}
									else {
										ni = (voxelObject -> resX) - 1;
									}

									if (jj < voxelObject -> resY) {
										nj = jj;
									}
									else {
										nj = (voxelObject -> resY) - 1;
									}
								
									if (kk < voxelObject -> resZ) {
										nk = kk;
									}
									else {
										nk = (voxelObject -> resX) - 1;
									}
									
									colorValue[index] = voxelObject -> voxelBuffer[ni][nj][nk];
									index++;
								}

							}
						}
						
						/* sort voxel data stream */
						qsort((void *)colorValue, 8, sizeof(colorValue[0]), SortFunction);

						/* compute color freq. */
						for(index = 0, tableIndex = 0; index < 8; ){
							freqTable[tableIndex][COLOR] = colorValue[index];
							while (freqTable[tableIndex][COLOR] == colorValue[index]){
								freqTable[tableIndex][FREQ]++;
								index++;
							}
							tableIndex++;
						}

						/* find max color freq. (color is not EMPTY_VOXEL) */
						maxFreqColor.color = EMPTY_VOXEL;
						maxFreqColor.freq  = 0;
						for(tableIndex = 0; tableIndex < 8; tableIndex++){
							if ((freqTable[tableIndex][FREQ] > maxFreqColor.freq) && (freqTable[tableIndex][COLOR]) != EMPTY_VOXEL){
								maxFreqColor.color = freqTable[tableIndex][COLOR];
								maxFreqColor.freq  = freqTable[tableIndex][FREQ];
							}
						}

						/* compute new color */
						if (freqTable[0][COLOR] == EMPTY_VOXEL) {
							if (freqTable[0][FREQ] > WEIGHT) {
								averageColor = (float)EMPTY_VOXEL;
							}
							else {
								averageColor = (float)maxFreqColor.color;
							}
						}
						else {
							averageColor = (float)maxFreqColor.color;
						}
						                
						newColor = (int)averageColor;
						voxelObjectLod1 -> voxelBuffer[i][j][k] = newColor;
					}
				} 
			}
			
			ReleaseVoxelObject(voxelObject);			
			voxelObject = CreateVoxelObject(objName, version, release,  voxelObjectLod1 -> resX, voxelObjectLod1 -> resY, voxelObjectLod1 -> resZ, LOD_1, cds, colorPalette);			
			for(k = 0; k < voxelObject -> resZ; k++){
				for(i = 0; i < voxelObject -> resX; i++){
					for(j = 0; j < voxelObject -> resY; j++){
						voxelObject -> voxelBuffer[i][j][k] = voxelObjectLod1 -> voxelBuffer[i][j][k];
					}
				}
			}
						
			break;
			
		case LOD_2:

			ComputeLOD(LOD_1, objName);
			printf("%s%d%s\n", "Computing level of detail (level ", nLod, ") ...");
			
			if (odd(voxelObject -> resX)) {
				newX = ((voxelObject -> resX) + 1) >> 1;
			}
			else {
				newX = (voxelObject -> resX) >> 1;
			}
			
			if (odd(voxelObject -> resY)) {
				newY = ((voxelObject -> resY) + 1) >> 1;
			}
			else {
				newY = (voxelObject -> resY) >> 1;
			}
			
			if (odd(voxelObject -> resZ)) {
				newZ = ((voxelObject -> resZ) + 1) >> 1;
			}
			else {
				newZ = (voxelObject -> resZ) >> 1;
			}
			
			voxelObjectLod2 = CreateVoxelObject(objName, version, release,  newX, newY, newZ, LOD_2, cds, colorPalette);

			printf("New resolution:\tx = %d\ty = %d\tz = %d\n", voxelObjectLod2 -> resX, voxelObjectLod2 -> resY, voxelObjectLod2 -> resZ);
			
			for(k = 0; k < voxelObjectLod2 -> resZ; k++){
				for(i = 0; i < voxelObjectLod2 -> resX; i++){
					for(j = 0; j < voxelObjectLod2 -> resY; j++){
					
						for (index = 0; index < 8; index++) {
							colorValue[index] = NaC;
							freqTable[index][COLOR] = NaC;
							freqTable[index][FREQ]  = 0;
						}						

						newColor = EMPTY_VOXEL;
						index = 0;
						tableIndex = 0;

						for (kk = 2*k; kk < (2*k + 2); kk++){
							for (ii = 2*i; ii < (2*i + 2); ii++){
								for (jj = 2*j; jj < (2*j + 2); jj++){

									if (ii < voxelObject -> resX) {
										ni = ii;
									}
									else {
										ni = (voxelObject -> resX) - 1;
									}

									if (jj < voxelObject -> resY) {
										nj = jj;
									}
									else {
										nj = (voxelObject -> resY) - 1;
									}
								
									if (kk < voxelObject -> resZ) {
										nk = kk;
									}
									else {
										nk = (voxelObject -> resX) - 1;
									}
									
									colorValue[index] = voxelObject -> voxelBuffer[ni][nj][nk];
									index++;
								}
							}
						}
						
						/* sort voxel data stream */
						qsort((void *)colorValue, 8, sizeof(colorValue[0]), SortFunction);

						/* compute color freq. */
						for(index = 0, tableIndex = 0; index < 8; ){
							freqTable[tableIndex][COLOR] = colorValue[index];
							while (freqTable[tableIndex][COLOR] == colorValue[index]){
								freqTable[tableIndex][FREQ]++;
								index++;
							}
							tableIndex++;
						}

						/* find max color freq. (color is not EMPTY_VOXEL) */
						maxFreqColor.color = EMPTY_VOXEL;
						maxFreqColor.freq  = 0;
						for(tableIndex = 0; tableIndex < 8; tableIndex++){
							if ((freqTable[tableIndex][FREQ] > maxFreqColor.freq) && (freqTable[tableIndex][COLOR]) != EMPTY_VOXEL){
								maxFreqColor.color = freqTable[tableIndex][COLOR];
								maxFreqColor.freq  = freqTable[tableIndex][FREQ];
							}
						}

						/* compute new color */
						if (freqTable[0][COLOR] == EMPTY_VOXEL) {
							if (freqTable[0][FREQ] > WEIGHT) {
								averageColor = (float)EMPTY_VOXEL;
							}
							else {
								averageColor = (float)maxFreqColor.color;
							}
						}
						else {
							averageColor = (float)maxFreqColor.color;
						}

						newColor = (int)averageColor;
						voxelObjectLod2 -> voxelBuffer[i][j][k] = newColor;
					}
				} 
			}
			
			ReleaseVoxelObject(voxelObject);			
			voxelObject = CreateVoxelObject(objName, version, release,  voxelObjectLod2 -> resX, voxelObjectLod2 -> resY, voxelObjectLod2 -> resZ, LOD_2, cds, colorPalette);
			for(k = 0; k < voxelObject -> resZ; k++){
				for(i = 0; i < voxelObject -> resX; i++){
					for(j = 0; j < voxelObject -> resY; j++){
						voxelObject -> voxelBuffer[i][j][k] = voxelObjectLod2 -> voxelBuffer[i][j][k];
					}
				}
			}
						
			break;
			
		case LOD_3:
		
			ComputeLOD(LOD_2, objName);
			printf("%s%d%s\n", "Computing level of detail (level ", nLod, ") ...");

			if (odd(voxelObject -> resX)) {
				newX = ((voxelObject -> resX) + 1) >> 1;
			}
			else {
				newX = (voxelObject -> resX) >> 1;
			}
			
			if (odd(voxelObject -> resY)) {
				newY = ((voxelObject -> resY) + 1) >> 1;
			}
			else {
				newY = (voxelObject -> resY) >> 1;
			}
			
			if (odd(voxelObject -> resZ)) {
				newZ = ((voxelObject -> resZ) + 1) >> 1;
			}
			else {
				newZ = (voxelObject -> resZ) >> 1;
			}
			
			voxelObjectLod3 = CreateVoxelObject(objName, version, release,  newX, newY, newZ, LOD_3, cds, colorPalette);

			printf("New resolution:\tx = %d\ty = %d\tz = %d\n", voxelObjectLod3 -> resX, voxelObjectLod3 -> resY, voxelObjectLod3 -> resZ);
			
			for(k = 0; k < voxelObjectLod3 -> resZ; k++){
				for(i = 0; i < voxelObjectLod3 -> resX; i++){
					for(j = 0; j < voxelObjectLod3 -> resY; j++){
					
						for (index = 0; index < 8; index++) {
							colorValue[index] = NaC;
							freqTable[index][COLOR] = NaC;
							freqTable[index][FREQ]  = 0;							
						}						

						newColor = EMPTY_VOXEL;
						index = 0;
						tableIndex = 0;
						
						for (kk = 2*k; kk < (2*k + 2); kk++){
							for (ii = 2*i; ii < (2*i + 2); ii++){
								for (jj = 2*j; jj < (2*j + 2); jj++){

									if (ii < voxelObject -> resX) {
										ni = ii;
									}
									else {
										ni = (voxelObject -> resX) - 1;
									}

									if (jj < voxelObject -> resY) {
										nj = jj;
									}
									else {
										nj = (voxelObject -> resY) - 1;
									}
								
									if (kk < voxelObject -> resZ) {
										nk = kk;
									}
									else {
										nk = (voxelObject -> resX) - 1;
									}
									
									colorValue[index] = voxelObject -> voxelBuffer[ni][nj][nk];
									index++;
								}
							}
						}
						
						/* sort voxel data stream */
						qsort((void *)colorValue, 8, sizeof(colorValue[0]), SortFunction);

						/* compute color freq. */
						for(index = 0, tableIndex = 0; index < 8; ){
							freqTable[tableIndex][COLOR] = colorValue[index];
							while (freqTable[tableIndex][COLOR] == colorValue[index]){
								freqTable[tableIndex][FREQ]++;
								index++;
							}
							tableIndex++;
						}

						/* find max color freq. (color is not EMPTY_VOXEL) */
						maxFreqColor.color = EMPTY_VOXEL;
						maxFreqColor.freq  = 0;
						for(tableIndex = 0; tableIndex < 8; tableIndex++){
							if ((freqTable[tableIndex][FREQ] > maxFreqColor.freq) && (freqTable[tableIndex][COLOR]) != EMPTY_VOXEL){
								maxFreqColor.color = freqTable[tableIndex][COLOR];
								maxFreqColor.freq  = freqTable[tableIndex][FREQ];
							}
						}

						/* compute new color */
						if (freqTable[0][COLOR] == EMPTY_VOXEL) {
							if (freqTable[0][FREQ] > WEIGHT) {
								averageColor = (float)EMPTY_VOXEL;
							}
							else {
								averageColor = (float)maxFreqColor.color;
							}
						}
						else {
							averageColor = (float)maxFreqColor.color;
						}
						                
						newColor = (int)averageColor;
						voxelObjectLod3 -> voxelBuffer[i][j][k] = newColor;
					}
				} 
			}
			
			ReleaseVoxelObject(voxelObject);			
			voxelObject = CreateVoxelObject(objName, version, release,  voxelObjectLod3 -> resX, voxelObjectLod3 -> resY, voxelObjectLod3 -> resZ, LOD_3, cds, colorPalette);
			for(k = 0; k < voxelObject -> resZ; k++){
				for(i = 0; i < voxelObject -> resX; i++){
					for(j = 0; j < voxelObject -> resY; j++){
						voxelObject -> voxelBuffer[i][j][k] = voxelObjectLod3 -> voxelBuffer[i][j][k];
					}
				}
			}
						
			break;
			
		case LOD_4:
			
			ComputeLOD(LOD_3, objName);
			printf("%s%d%s\n", "Computing level of detail (level ", nLod, ") ...");

			if (odd(voxelObject -> resX)) {
				newX = ((voxelObject -> resX) + 1) >> 1;
			}
			else {
				newX = (voxelObject -> resX) >> 1;
			}
			
			if (odd(voxelObject -> resY)) {
				newY = ((voxelObject -> resY) + 1) >> 1;
			}
			else {
				newY = (voxelObject -> resY) >> 1;
			}
			
			if (odd(voxelObject -> resZ)) {
				newZ = ((voxelObject -> resZ) + 1) >> 1;
			}
			else {
				newZ = (voxelObject -> resZ) >> 1;
			}
			
			voxelObjectLod4 = CreateVoxelObject(objName, version, release,  newX, newY, newZ, LOD_4, cds, colorPalette);

			printf("New resolution:\tx = %d\ty = %d\tz = %d\n", voxelObjectLod4 -> resX, voxelObjectLod4 -> resY, voxelObjectLod4 -> resZ);
			
			for(k = 0; k < voxelObjectLod4 -> resZ; k++){
				for(i = 0; i < voxelObjectLod4 -> resX; i++){
					for(j = 0; j < voxelObjectLod4 -> resY; j++){
					
						for (index = 0; index < 8; index++) {
							colorValue[index] = NaC;
							freqTable[index][COLOR] = NaC;
							freqTable[index][FREQ]  = 0;														
						}						

						newColor = EMPTY_VOXEL;
						index = 0;
						tableIndex = 0;
												
						
						for (kk = 2*k; kk < (2*k + 2); kk++){
							for (ii = 2*i; ii < (2*i + 2); ii++){
								for (jj = 2*j; jj < (2*j + 2); jj++){

									if (ii < voxelObject -> resX) {
										ni = ii;
									}
									else {
										ni = (voxelObject -> resX) - 1;
									}

									if (jj < voxelObject -> resY) {
										nj = jj;
									}
									else {
										nj = (voxelObject -> resY) - 1;
									}
								
									if (kk < voxelObject -> resZ) {
										nk = kk;
									}
									else {
										nk = (voxelObject -> resX) - 1;
									}
									
									colorValue[index] = voxelObject -> voxelBuffer[ni][nj][nk];
									index++;
								}
							}
						}

						/* sort voxel data stream */
						qsort((void *)colorValue, 8, sizeof(colorValue[0]), SortFunction);

						/* compute color freq. */
						for(index = 0, tableIndex = 0; index < 8; ){
							freqTable[tableIndex][COLOR] = colorValue[index];
							while (freqTable[tableIndex][COLOR] == colorValue[index]){
								freqTable[tableIndex][FREQ]++;
								index++;
							}
							tableIndex++;
						}

						/* find max color freq. (color is not EMPTY_VOXEL) */
						maxFreqColor.color = EMPTY_VOXEL;
						maxFreqColor.freq  = 0;
						for(tableIndex = 0; tableIndex < 8; tableIndex++){
							if ((freqTable[tableIndex][FREQ] > maxFreqColor.freq) && (freqTable[tableIndex][COLOR]) != EMPTY_VOXEL){
								maxFreqColor.color = freqTable[tableIndex][COLOR];
								maxFreqColor.freq  = freqTable[tableIndex][FREQ];
							}
						}

						/* compute new color */
						if (freqTable[0][COLOR] == EMPTY_VOXEL) {
							if (freqTable[0][FREQ] > WEIGHT) {
								averageColor = (float)EMPTY_VOXEL;
							}
							else {
								averageColor = (float)maxFreqColor.color;
							}
						}
						else {
							averageColor = (float)maxFreqColor.color;
						}
						                
						newColor = (int)averageColor;
						voxelObjectLod4 -> voxelBuffer[i][j][k] = newColor;
					}
				} 
			}
			
			ReleaseVoxelObject(voxelObject);			
			voxelObject = CreateVoxelObject(objName, version, release,  voxelObjectLod4 -> resX, voxelObjectLod4 -> resY, voxelObjectLod4 -> resZ, LOD_4, cds, colorPalette);
			for(k = 0; k < voxelObject -> resZ; k++){
				for(i = 0; i < voxelObject -> resX; i++){
					for(j = 0; j < voxelObject -> resY; j++){
						voxelObject -> voxelBuffer[i][j][k] = voxelObjectLod4 -> voxelBuffer[i][j][k];
					}
				}
			}
						
			break;
			
		default:
		
			break;
	}
	return;
}


/*                                                            */
/*  Convert voxel object to solid faces (B-rep)               */
/*                                                            */
GLuint ConvertVoxelObject2Faces(char *sourceFile, LOD newLod, int flag){

	register Index index = 0;
	register Counter i, j , k;
	GLuint aList;


	maxNumOfInputQuads   = 0;
	realNumOfInputQuads  = 0;
	realNumOfOutputQuads = 0;
	ratio    = 0.0;
	palIndex = 1;	
	
	if ((textureFile = fopen(sourceFile, "r")) == NULL){
		CatchException;
	}

	#if defined(DEBUG)
		if ((faceFile = fopen(DEBUG_FILE, "w")) == NULL){
			CatchException;
		}
	#endif

	if ((timeStartInterval = (time_t *) malloc(sizeof(time_t))) == NULL){
		CatchException;
	}

	if ((timeEndInterval = (time_t *) malloc(sizeof(time_t))) == NULL){
		CatchException;
	}

	for(index = 0; index < PALETTE_SIZE; index++){
		colorPalette[index].red   = 0;
		colorPalette[index].green = 0;
		colorPalette[index].blue  = 0;
	}

	startTime = time(timeStartInterval);
	
	ParseTextureFile();
	
	endTime = time(timeEndInterval);
	
	printf("%s%3d[s]\n", "Time (reading data): ", (int)difftime(endTime, startTime));

	printf("%s\n", "Creating voxel object  ...");
	
	startTime = time(timeStartInterval);

	voxelObject = CreateVoxelObject(sourceFile, version, release,  x, y, z, LOD_0, cds, colorPalette);

	for(k = 0; k < z; k++){
		for(i = 0; i < x; i++){
			for(j = 0; j < y; j++){
				voxelObject -> voxelBuffer[i][j][k] = voxelBuffer[i][j][k];
			}
		} 
	}

	endTime = time(timeEndInterval);
	
	printf("%s%3d[s]\n", "Time (creating object): ", (int)difftime(endTime, startTime));

	printf("Object name: %s\n", voxelObject -> objectName);

	#if defined(DEBUG)
		printf("Version: %d\n", voxelObject -> version);
		printf("Release: %d\n", voxelObject -> release);
		fflush(stdout);
	#endif
	
	printf("Resolution:\tx = %d\ty = %d\tz = %d\n", voxelObject -> resX, voxelObject -> resY, voxelObject -> resZ);

	printf("Original level of detail: %d\n", voxelObject -> lod);

	if (voxelObject -> cdsFlag == 0) {
		#if defined(DEBUG)
			printf("%s\n", "Non-compressed data stream");
		#endif
	}
	else{
		printf("%s\n", "Compressed data stream");
		printf("%s\n", "Sorry, copressed data stream not supported in this version:");
		printf("%s\nexiting ...\n", VGE_VERSION);
		ReleaseVoxelObject(voxelObject);
		exit(ERROR);
	}
	
	#if defined(DEBUG)
	printf("%s\n", "Color palette:");
	for(index = 0; index < palIndex; index++){
		printf("c[%03d] %3d %3d %3d\n", index, voxelObject -> colorPalette[index].red, voxelObject -> colorPalette[index].green, voxelObject -> colorPalette[index].blue);
	}
	#endif
	
	startTime = time(timeStartInterval);
	
	ComputeLOD(newLod, sourceFile);
	
	maxNumOfInputQuads = ((voxelObject -> resX) * (voxelObject -> resY) * (voxelObject -> resZ) * 6);
	
	printf("%s", "Searching for quads, please, stand by ... ");
	
	fflush(stdout);
	
	for(k = 0; k < voxelObject -> resZ; k++){
		for(i = 0; i < voxelObject -> resX; i++){
			for(j = 0; j < voxelObject -> resY; j++){
				if ((voxelObject -> voxelBuffer[i][j][k]) != EMPTY_VOXEL) {
					realNumOfInputQuads += 	6;
				}
			}
		} 
	}		
	
	aList = glGenLists(1);
        glNewList(aList, GL_COMPILE);
	  glEnable(GL_DEPTH_TEST);
	  glEnable(GL_CULL_FACE);
	  glPushMatrix();
	  glRotatef(-90.0, 1.0, 0.0, 0.0);
	  glRotatef(-90.0, 0.0, 0.0, 1.0);
	  Voxel2Faces(DIRECTION_X);
	  Voxel2Faces(DIRECTION_Y);
	  Voxel2Faces(DIRECTION_Z);
	  glPopMatrix();
	glDisable (GL_CULL_FACE);
	glEndList();
	
	endTime = time(timeEndInterval);
	
	printf("%s\n", "done");
	
	switch (newLod){
	
		case LOD_1: 
	
			ReleaseVoxelObject(voxelObjectLod1);
			
			break;
		
		case LOD_2:
	
			ReleaseVoxelObject(voxelObjectLod1);
			ReleaseVoxelObject(voxelObjectLod2);
			
			break;
			
		case LOD_3:
	
			ReleaseVoxelObject(voxelObjectLod1);
			ReleaseVoxelObject(voxelObjectLod2);
			ReleaseVoxelObject(voxelObjectLod3);
			
			break;
			
		case LOD_4:
	
			ReleaseVoxelObject(voxelObjectLod1);
			ReleaseVoxelObject(voxelObjectLod2);
			ReleaseVoxelObject(voxelObjectLod3);
			ReleaseVoxelObject(voxelObjectLod4);
			
			break;
			
		default:
		
			break;
	}
	
	ReleaseVoxelObject(voxelObject);
	
	fclose(textureFile);
	#if defined(DEBUG)
		fclose(faceFile);
	#endif	

	printf("%s% 12d\n", "Number of all quads (solid + transparent) in input: ", maxNumOfInputQuads);
	printf("%s\t    % 12d\n", "Number of quads (solid only) in input:  ", realNumOfInputQuads);
	printf("%s\t    % 12d\n", "Number of quads (solid only) in output: ", realNumOfOutputQuads); 
	
	if (realNumOfInputQuads != 0.0) {
		ratio = ((((Ratio)realNumOfOutputQuads) / ((Ratio)realNumOfInputQuads)) * 100.0);
	}
	else {
		ratio = 100.0;
	}
	
	printf("%s%.5f[%c]\n%s%3d[s]\n", "Ratio: ", ratio, '%', "Time (computing LOD + searching quads): ", (int)difftime(endTime, startTime));
	
	#if defined(DEBUG)
		printf("%s %s\n", "Output writen on" , DEBUG_FILE);
	#endif
	
	fflush(stdout);
		
	return aList;		
}

/* end of file vge.c */

/* this was added by J.Z. */


void parseTextureFile(FILE *tF){

	register Index i, j, k;


	while (!feof(tF)){
	
		tmpChar = fgetc(tF);

		switch(tmpChar){
		
			case '/': 
			
				do {
					tmpChar = fgetc(tF);
				} while (tmpChar != '/');
				tmpChar = fgetc(tF);
				parseTextureFile(tF);
				
				break;

			case '#':
			
				fgets(line, LINE_SIZE, tF);
				sscanf(line, "%3s", &token);

				if (!strcmp(token, "vtf")) {
					printf("%s\n", "Voxel texture format");
					printf("%s\n", "Reading voxel data sream, please, stand by ...");
					fflush(stdout);
				}

				if (!strcmp(token, "ver")) {
					sscanf(line, "%s %d.%d", &token, &version, &release);

				}

				if (!strcmp(token, "res")) {
					sscanf(line, "%s %d %d %d", &token, &x, &y, &z);
					return;
				}
	
				if (!strcmp(token, "lod")) {
					sscanf(line, "%s %d", &token, &lod);
					
				}

				if (!strcmp(token, "cds")) {
					sscanf(line, "%s %d", &token, &cds);
				}
				
				if (!strcmp(token, "vds")) {
					voxel = fgetc(tF);
					if (voxel == '/') {
						do {
							voxel = fgetc(tF);
						} while (voxel != '/');
						voxel = fgetc(tF);
					}
					else {
						ungetc(voxel, tF);
					}

					voxelBuffer = CreateVoxelBuffer(x, y, z);
					for(k = 0; k < z; k++){
						for(i = 0; i < x; i++){
							for(j = 0; j < y; j++){
								fscanf(tF, "%d", &voxel);
								voxelBuffer[i][j][k] = voxel;
							}
						} 
					}		
				}
				
				parseTextureFile(tF);
				
				break;

			case 'c':
			
				if (palIndex < PALETTE_SIZE) {
					fscanf(tF, "%d %d %d", &colorPalette[palIndex].red, &colorPalette[palIndex].green, &colorPalette[palIndex].blue);
					palIndex++;
				}
				parseTextureFile(tF);
				
				break;
				
			default:
			
				break;
		}
	}
	return;
}




void getTextureSize(char *fn, Resolution *sx, Resolution *sy, Resolution *sz){
    FILE *tF;

    fprintf(stderr, "Relolution of %s\n", fn);
    if ((tF = fopen(fn, "r")) == NULL){
	CatchException;
    }
    parseTextureFile(tF);
    fclose(tF);
    
    *sx = x;
    *sy = y;
    *sz = z;
    return;
}
