#include <stdio.h>
#include <GL/gl.h>
#include "viz.h"
#include <GL/glut.h>
#include "mui/mui.h"

extern TexObj *Objs;
extern int NumObjs;
extern int View;

int writeScene (const char *fileName){
    FILE *out;
    int i;
    glutSetWindow (View);
    out = fopen (fileName, "wb");
    if (out==NULL){
	fprintf (stderr, "error writing file (%s)\n", fileName);
	return 1;
    }
    fwrite (&NumObjs, sizeof(int), 1, out);
    for (i=0; i<NumObjs; i++){
	fwrite (Objs+i, sizeof(TexObj), 1, out);
    }
    /* now write landscape info */
    fwrite (&Level, sizeof(int), 1, out);
    fwrite (Height[0], sizeof(GLfloat), 3 * Dim(Level) * Dim(Level), out);
    fclose (out);
    return 0;
}

extern void FreeMemory (void);
extern void AllocMemory (void);

void freeGLLists (){
    int i, j;
    
    for (i=0; i<NumObjs; i++){
	for (j=0; j<4; j++){
	    glDeleteLists (Objs[i].list[j], 1);
	}
    }
}
extern int ConvertVoxelObject2Faces (char *, int, int);
int readScene (const char *fileName){
    FILE *in;
    int i, j, lastValidLOD;
    char fName[256];
    
    glutSetWindow (View);
    in = fopen (fileName, "rb");
    if (in == NULL){
	fprintf (stderr, "error opening file (%s)\n", fileName);
	return 1;
    }
    freeGLLists();		       /* Objects */
    free (Objs);
    NumObjs = 0;
    Objs = NULL;
    FreeMemory();		       /* landscape */
    fread (&NumObjs, sizeof(int), 1, in);
    if (NumObjs>0){
	Objs = (TexObj*)malloc (sizeof(TexObj) * NumObjs);
	fread (Objs, sizeof(TexObj), NumObjs, in);
    }
    fread (&Level, sizeof(int), 1, in);
    AllocMemory();		       /* landscape */
    fread (Height[0], sizeof(GLfloat), 3 * Dim(Level) * Dim(Level), in);
    fclose (in);
    /* all is read, now make new gl lists for objects */
    for (i=0; i<NumObjs; i++){
	printf ("Object no. %i\n", i);
	for (j=0; j<4; j++){
	    if (!Objs[i].t.isVRMLLOD){
		strcpy (fName, Objs[i].t.path);
		strcat (fName, Objs[i].t.fName[j]);
		Objs[i].list[j] = ConvertVoxelObject2Faces ( fName, j, 0);
	    } else {
		strcpy (fName, Objs[i].t.path);
		if (strcmp (Objs[i].t.fName[j], "")){
		    strcat (fName, Objs[i].t.fName[j]);
		    Objs[i].list[j] = ConvertVoxelObject2Faces ( fName, 0, 0);
		    lastValidLOD = j;
		} else {
		    if (j==0){
			fprintf (stderr, "error : no LOD 0 defined\n");
			exit (1);
		    }
		    strcat (fName, Objs[i].t.fName[lastValidLOD]);
		    Objs[i].list[j] = ConvertVoxelObject2Faces ( fName, 0, 0);
		}
		printf ("VRMLLOD\n");
	    }
	}
    }
    landChanged();
    return 0;
}

static char fileName[256] = "";
static muiObject *line;
static muiObject *button;
static int fdWin = -1;
extern int win;
static void fdClick (muiObject *o, enum muiReturnValue v){
    switch (muiGetID(o)){
     case -1:
	break;
     case 0:			       /* load */
	printf("Load\n");
	strcpy (fileName, muiGetTBString (line));
	readScene (fileName);
	break;
     case 1:			       /* save */
	printf("Save\n");
	strcpy (fileName, muiGetTBString (line));
	writeScene (fileName);
	break;
	
    }
    printf ("%s\n", fileName);
    glutSetWindow (fdWin);
    glutHideWindow ();
    glutSetWindow (win);
}
void fileDialog (int ls){

    muiObject *o;
    if (fdWin<0){
	glutInitWindowPosition (200, 250);
	glutInitWindowSize (240, 95);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
	fdWin = glutCreateWindow (ls==0 ? "Load" : "Save");
	muiNewUIList (4);
	muiInit (4);
	line = muiNewTextbox ( 20, 220, 50);
	button = muiNewButton ( 50, 100, 10, 35);
	o = muiNewButton ( 140, 190, 10, 35);
	muiLoadButton (o, "Cancel");
	muiSetID (o, -1);
	muiSetCallback (o, fdClick);
    } else {
	glutSetWindow (fdWin);
    }
   
    muiSetTBString (line, fileName);
    muiLoadButton (button, ls==0 ? "Load":"Save");
    muiSetID (button, ls);
    muiSetCallback (button, fdClick);
    glutSetWindowTitle (ls==0 ? "Load" : "Save");
    glutShowWindow ();

    glutPostRedisplay();
}
