#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dirent.h>
#include <math.h>
#include <limits.h>           /* ULONG_MAX is defined here */
#include <float.h>            /* FLT_MAX is atleast defined here */

#include <time.h>  /* for random seed */

#include "landviewer.h"


int Level;
GLfloat **Height;

float FractalStep = 0.5;
/***************************************************************/
/************************* VECTOR JUNK *************************/
/***************************************************************/

  /* normalizes v */
void normalize(GLfloat v[3])
{
  GLfloat d = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);

  if (d == 0)
    fprintf(stderr, "Zero length vector in normalize\n");
  else
    v[0] /= d; v[1] /= d; v[2] /= d;
}

  /* calculates a normalized crossproduct to v1, v2 */
void ncrossprod(float v1[3], float v2[3], float cp[3])
{
  cp[0] = v1[1]*v2[2] - v1[2]*v2[1];
  cp[1] = v1[2]*v2[0] - v1[0]*v2[2];
  cp[2] = v1[0]*v2[1] - v1[1]*v2[0];
  normalize(cp);
}

  /* calculates normal to the triangle designated by v1, v2, v3 */
void triagnormal(float v1[3], float v2[3], float v3[3], float norm[3])
{
  float vec1[3], vec2[3];

  vec1[0] = v3[0] - v1[0];  vec2[0] = v2[0] - v1[0];
  vec1[1] = v3[1] - v1[1];  vec2[1] = v2[1] - v1[1];
  vec1[2] = v3[2] - v1[2];  vec2[2] = v2[2] - v1[2];

  ncrossprod(vec2, vec1, norm);
}
/* calculate middle normal  */
void midNormal (float n1[3], float n2[3], float o[3]){
  o[0] = (n1[0] + n2[0])/2;
  o[1] = (n1[1] + n2[1])/2;
  o[2] = (n1[2] + n2[2])/2;
}

float xzlength(float v1[3], float v2[3])
{
  return sqrt((v1[0] - v2[0])*(v1[0] - v2[0]) +
              (v1[2] - v2[2])*(v1[2] - v2[2]));
}

float xzslope(float v1[3], float v2[3])
{
  return ((v1[0] != v2[0]) ? ((v1[2] - v2[2]) / (v1[0] - v2[0]))
	                   : FLT_MAX);
}

/***************************************************************/
/************************ MOUNTAIN STUFF ***********************/
/***************************************************************/
int Dim(int level)
{
   return 1<<(level);
}

int IndexOf(int x,
	    int y,
	    int level)
{
   return x + y * Dim(level);
}
void AllocMemory( void)
{
    int j;
    GLfloat *tmp;
    
    Height = (GLfloat**) malloc( sizeof(GLfloat*) * Dim(Level) * Dim(Level));
    tmp = (GLfloat*)malloc (sizeof(GLfloat) *Dim(Level)*Dim(Level)*3);
    for (j=0; j<Dim(Level)*Dim(Level); j++)
      Height[j] = tmp + 3*j;
}
void FreeMemory( void)
{
   int j;
   return;
   for( j=0; j<Dim(Level)*Dim(Level); j++)
     free( Height[j]);
   free( Height);
   Height = NULL;
}

static float FlatFunction( int x, int y, int maxXY)
{
   return 0.5;
}

float **FractalHeights;

#define RAND_COEFFICIENT 0.1
static float FractalFunction( int x, int y, int maxXY)
{
    if( FractalHeights !=NULL){
	if( x==maxXY-1 && y==maxXY-1){	       /* last acces -> dealloc memory */
	    float tmp;
	    int xi;
	    
	    tmp = FractalHeights[x][y];
	    for( xi=0; xi<Dim(Level)+1; xi++)
	      free( FractalHeights[xi]);
	    free( FractalHeights);
	    FractalHeights = NULL;
	    return tmp;
	}
	return FractalHeights[x][y];
    } else {			       /* generate new fractal */
	int xi, yi, level;
	int displacement;
	float step;
	
	step = FractalStep;
	FractalHeights = (float**) malloc( sizeof(float) * (Dim(Level)+1));
	for( xi=0; xi<=Dim(Level); xi++)
	  FractalHeights[xi] = (float*) malloc( sizeof(float) * (Dim(Level)+1));

	FractalHeights[0][0] = FractalHeights[0][Dim(Level)] =
	  FractalHeights[Dim(Level)][0] = FractalHeights[Dim(Level)][Dim(Level)] = 0.5;
	
	for( level=1; level<=Level; level++){
	    displacement = 1 << (Level - level + 1);
	    for( yi=0; yi<=Dim(Level); yi+=displacement)
	      for( xi=0; xi<Dim(Level); xi+=displacement){
		  FractalHeights[xi+displacement/2][yi] =
		    (FractalHeights[xi][yi] + FractalHeights[xi+displacement][yi]) / 2.0 +
		    step * ((float)rand()/(float)RAND_MAX - 0.5);
		  if( FractalHeights[xi+displacement/2][yi]>1.0)
		    FractalHeights[xi+displacement/2][yi] = 1.0;
		  if( FractalHeights[xi+displacement/2][yi]<0.0)
		    FractalHeights[xi+displacement/2][yi] = 0.0;
	      }
	    for( xi=0; xi<=Dim(Level); xi+=displacement/2)
	      for( yi=0; yi<Dim(Level); yi+=displacement){
		  FractalHeights[xi][yi+displacement/2] =
		    (FractalHeights[xi][yi] + FractalHeights[xi][yi+displacement]) / 2.0 +
		    step * ((float)rand()/(float)RAND_MAX - 0.5) ;
		  if( FractalHeights[xi][yi+displacement/2] > 1.0)
		    FractalHeights[xi][yi+displacement/2] = 1.0;
		  if( FractalHeights[xi][yi+displacement/2] < 0.0)
		    FractalHeights[xi][yi+displacement/2] = 0.0;
	      }
	    step /= 2;
	}				       /* level */
	return FractalHeights[x][y];
    }
}

/* create GL viewing list of area  */

static void countNormal (int x, int y, float norm[3]){
  float norms[6][3];
  int i;
  triagnormal(Height[IndexOf( x-1, y-1, Level)], 
	      Height[IndexOf( x, y-1, Level)], 
	      Height[IndexOf( x, y, Level)],
	      norms[0]);
  triagnormal(Height[IndexOf( x-1, y-1, Level)], 
	      Height[IndexOf( x, y, Level)], 
	      Height[IndexOf( x-1, y, Level)],
	      norms[1]);
  triagnormal(Height[IndexOf( x, y, Level)], 
	      Height[IndexOf( x, y+1, Level)], 
	      Height[IndexOf( x-1, y, Level)],
	      norms[2]);
  triagnormal(Height[IndexOf( x, y, Level)], 
	      Height[IndexOf( x, y+1, Level)], 
	      Height[IndexOf( x+1, y+1, Level)],
	      norms[3]);
  triagnormal(Height[IndexOf( x, y, Level)], 
	      Height[IndexOf( x+1, y+1, Level)], 
	      Height[IndexOf( x+1, y, Level)],
	      norms[4]);
  triagnormal(Height[IndexOf( x, y, Level)], 
	      Height[IndexOf( x+1, y, Level)], 
	      Height[IndexOf( x, y-1, Level)],
	      norms[5]);
  midNormal (norms[0], norms[1], norm);
  for (i=2; i<6; i++){
    midNormal (norm, norms[i], norm);
  }
}
void CreateLandList(int list, int material)
{
    int x, y;
    float norm[3];
    int i;

    glNewList(list, GL_COMPILE);
    glPushAttrib(GL_LIGHTING_BIT);
    glCallList(material);
    y = 0;
    for( x=1; x<Dim(Level)-2; x++){
	y = 1;
	glBegin(GL_TRIANGLE_STRIP);
	countNormal (x+1, y, norm);
	glNormal3fv(norm);
	glVertex3fv(Height[IndexOf( x+1, y, Level)]);

	countNormal (x, y, norm);
	glNormal3fv(norm);
	glVertex3fv(Height[IndexOf( x, y, Level)]);

	for( y=2; y<Dim(Level)-2; y++){
	  countNormal (x+1, y, norm);
	  glNormal3fv(norm);
	  glVertex3fv(Height[IndexOf( x+1, y, Level)]);
	    
	  countNormal (x, y, norm);
	  glNormal3fv(norm);
	  glVertex3fv(Height[IndexOf( x, y, Level)]);
	}
	glEnd();
    }
    glPopAttrib();
    glEndList();
    return;
}

static void NewLand( float (*landFunc)( int x, int y, int maxXY))
{
   int x, y;
   
   for( x=0; x<Dim(Level); x++)
     for( y=0; y<Dim(Level); y++){
	Height[IndexOf( x, y, Level)][0] = ((float)x / Dim(Level) * 2.0 ) - 1.0;
	Height[IndexOf( x, y, Level)][1] = landFunc( x, y, Dim(Level));
	Height[IndexOf( x, y, Level)][2] = ((float) y / Dim(Level) * 2.0) - 1.0;
     };
   return;
}
/***************************************************************/
/**************************** MAIN *****************************/
/***************************************************************/

void InitLand(int level){
    Level = level;
    AllocMemory();
    random();
    NewLand( FractalFunction);
}
void Raise (float x, float z, int dist){
    int d;
    d = Dim(Level);
    Height[IndexOf((int) (((x+1.0)/2.0)*(float)d),
		   (int) (((z+1.0)/2.0)*(float)d), Level)][1] += dist * 0.005; 
    landChanged ();
}
void RaiseRegion (float x1, float z1, float x2, float z2, int dist){
    int ix1, ix2, iz1, iz2, tmp;
    float fdist;
    int i, j;
    int d;
    d = Dim(Level);
    fdist = dist * 0.005;
    ix1 = (int) floor(((x1+1.0)/2.0)*(float)d);
    ix2 = (int) floor(((x2+1.0)/2.0)*(float)d);
    iz1 = (int) floor(((z1+1.0)/2.0)*(float)d);
    iz2 = (int) floor(((z2+1.0)/2.0)*(float)d);
    if (ix1>ix2){
	tmp = ix1;
	ix1 = ix2;
	ix2 = tmp;
    }
    if (iz1>iz2){
	tmp = iz1;
	iz1 = iz2;
	iz2 = tmp;
    }
    ix2++; iz2++;
    if (iz1< 0) iz1=0;
    if (iz2>=Dim(Level)) iz2 = Dim(Level) - 1;
    if (ix1<=0) ix1 = 0;
    if (ix2>=Dim(Level)) ix2 = Dim(Level)-1;
    
    for (i=ix1; i<ix2; i++)
      for (j=iz1; j<iz2; j++){
	     Height[IndexOf(i, j , Level)][1] += fdist;
      }
    landChanged();
}
void RaiseRegionTo (float x1, float z1, float x2, float z2, float h){
    int ix1, ix2, iz1, iz2, tmp;
    int i, j;
    int d;
    d = Dim(Level);
    ix1 = (int) floor(((x1+1.0)/2.0)*(float)d);
    ix2 = (int) floor(((x2+1.0)/2.0)*(float)d);
    iz1 = (int) floor(((z1+1.0)/2.0)*(float)d);
    iz2 = (int) floor(((z2+1.0)/2.0)*(float)d);
    if (ix1>ix2){
	tmp = ix1;
	ix1 = ix2;
	ix2 = tmp;
    }
    if (iz1>iz2){
	tmp = iz1;
	iz1 = iz2;
	iz2 = tmp;
    }
    ix2++;
    iz2++;
    printf("%i, %i - %i, %i\n", ix1, iz1, ix2, iz2);
    if (iz1< 0) iz1=0;
    if (iz2>=Dim(Level)) iz2 = Dim(Level) - 1;
    if (ix1<=0) ix1 = 0;
    if (ix2>=Dim(Level)) ix2 = Dim(Level)-1;
    for (i=ix1; i<ix2; i++)
      for (j=iz1; j<iz2; j++){
	     Height[IndexOf(i, j , Level)][1] = h+0.5;
      }
    landChanged();
}
void newLand (int i){
    switch (i){
     case 0:
	NewLand (FlatFunction);
	break;
     case 1:
	FractalStep = 0.4;
	NewLand (FractalFunction);
	break;
     case 2:
	FractalStep = 0.9;
	NewLand (FractalFunction);
	break;
     case 3:
	FractalStep = 1.4;
	NewLand (FractalFunction);
	break;
    }
    landChanged();
}

void landSize (int maxX){
    FreeMemory();
    switch (maxX){
     case 16:
	Level = 4;
	break;
     case 32:
	Level = 5;
	break;
     case 64:
	Level = 6;
	break;
     case 128:
	Level = 7;
	break;
    }
    AllocMemory();
    NewLand (FractalFunction);
    landChanged();
}
float heightAt (float x, float y){
    int d, ix, iy;
    float bx1, bx2, by;
    float h1, h2;
    
    if (x<-1.0 || y<-1.0 || x>1.0 || y>1.0)
      return 0.0;
    d = Dim (Level);
    ix = (int) floor(((x+1.0)/2.0)*(float)d);
    iy = (int) floor(((y+1.0)/2.0)*(float)d);
    if (ix>=Dim(Level)-1 ||
	iy >= Dim(Level)-1)
      return Height[IndexOf (ix,iy, Level)][1] - 0.45;
    /* bi-linear interpolation */
    bx1 = 2.0 * (float)ix / (float)(d-1) - 1.0;
    bx2 = 2.0 * (float)(ix+1) / (float)(d-1) - 1.0;
    by = 2.0 * (float)(iy) / (float)(d-1) - 1.0;
    h1 = Height[IndexOf(ix, iy , Level)][1] +
      (x-bx1) * ( Height[IndexOf(ix+1, iy , Level)][1] -  Height[IndexOf(ix, iy , Level)][1]);
    h2 = Height[IndexOf(ix+1, iy+1 , Level)][1] +
      (x-bx2) * ( Height[IndexOf(ix+1, iy+1 , Level)][1] -  Height[IndexOf(ix, iy+1 , Level)][1]);
    return (h1 + (y-by) * (h2 - h1)) - 0.45;
	    
}
