#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <GL/glut.h>
#include <mui/mui.h>
#include "viz.h"
#include "landviewer.h"

static muiObject *ModeLabel;
static muiObject *ExtraButtons[4];

static void hideExtraButtons (){
    int i;
    for (i=0; i<4; i++) muiSetEnable (ExtraButtons[i], 0);
    muiLoadButton (ExtraButtons[0], "");
    muiLoadButton (ExtraButtons[1], "");
    muiLoadButton (ExtraButtons[2], "");
    muiLoadButton (ExtraButtons[3], "");
}
static void showViewButtons (){
    int i;
    for (i=0; i<4; i++) muiSetEnable (ExtraButtons[i], 1);
    muiSetEnable (ExtraButtons[3], 0);
    muiLoadButton (ExtraButtons[0], "Polar");
    muiSetID (ExtraButtons[0], MENU_POLAR);
    muiLoadButton (ExtraButtons[1], "Fly");
    muiSetID (ExtraButtons[1], MENU_FLY);
    muiLoadButton (ExtraButtons[2], "Walk");
    muiSetID (ExtraButtons[2], MENU_WALK);
    muiLoadButton (ExtraButtons[3], "");
    glutPostRedisplay();
}
static void showEditButtons (){
    int i;
    for (i=0; i<4; i++) muiSetEnable (ExtraButtons[i], 1);
    muiLoadButton (ExtraButtons[0], "Move");
    muiSetID (ExtraButtons[0], MENU_MOVE_OBJECT);
    muiLoadButton (ExtraButtons[1], "Delete");
    muiSetID (ExtraButtons[1], MENU_DELETE_OBJECT);
    muiLoadButton (ExtraButtons[2], "Fall to gnd");
    muiSetID (ExtraButtons[2], MENU_FALL_OBJECT);
    muiLoadButton (ExtraButtons[3], "Raise under");
    muiSetID (ExtraButtons[3], MENU_RAISE_OBJECT);
    glutPostRedisplay();
}
static void showLandButtons (){
    muiSetEnable (ExtraButtons[0], 1);
     muiLoadButton (ExtraButtons[0], "Level");
    muiSetID (ExtraButtons[0], MENU_LEVEL);
    glutPostRedisplay();
}

void menuFunc (int value){
    switch (value){
     case MENU_QUIT:
	exit (0);
     case MENU_LOAD:
	fileDialog (0);
	break;
     case MENU_SAVE:
	fileDialog (1);
	break;
     case MENU_16:
	landSize (16);
	break;
     case MENU_32:
	landSize (32);
	break;
     case MENU_64:
	landSize (64);
	break;
     case MENU_128:
	landSize (128);
	break;
     case MENU_FLAT:
	newLand (0);
	break;
     case MENU_FRACT_1:
	newLand (1);
	break;
     case MENU_FRACT_2:
	newLand (2);
	break;
     case MENU_FRACT_3:
	newLand (3);
	break;
     case MENU_LEVEL:
	levelRegion();
	break;
     case MENU_POLAR:
	muiChangeLabel (ModeLabel, "Mode: view(polar)");
	hideExtraButtons ();
	showViewButtons ();
	setMode (MENU_POLAR);
	break;
     case MENU_FLY:
	muiChangeLabel (ModeLabel, "Mode: view(fly)");
	hideExtraButtons ();
	showViewButtons ();
	setMode (MENU_FLY);
	break;
     case MENU_WALK:
	muiChangeLabel (ModeLabel, "Mode: view(walk)");
	hideExtraButtons ();
	showViewButtons ();
	setMode (MENU_WALK);
	break;
     case MENU_INSERT_OBJECT:
	muiChangeLabel (ModeLabel, "Mode: insert object");
	hideExtraButtons ();
	setMode (MENU_INSERT_OBJECT);
	break;
     case MENU_DELETE_OBJECT:
	deleteObject ();
	break;
     case MENU_FALL_OBJECT:
	fallObject ();
	break;
     case MENU_RAISE_OBJECT:
	raiseObject ();
	break;
     case MENU_EDIT_LAND:
	muiChangeLabel (ModeLabel, "Mode: edit land");
	hideExtraButtons ();
	showLandButtons ();
	setMode (MENU_EDIT_LAND);
	break;
     case MENU_EDIT_OBJECT:
	muiChangeLabel (ModeLabel, "Mode: edit object");
	hideExtraButtons ();
	showEditButtons ();
	setMode (MENU_EDIT_OBJECT);
	break;
     case MENU_MOVE_OBJECT:
	muiChangeLabel (ModeLabel, "Mode: move object");
	setMode (MENU_MOVE_OBJECT);
	hideExtraButtons ();
	break;
     default:
	printf ("Function not supported\n");
    }
}
static void buttonFunc (muiObject *b, enum muiReturnValue v){
    menuFunc (muiGetID (b));
}

static muiObject *makeButton (int x, int y, int width, int height, 
			      int id, char *name,
			      void (*callback)(muiObject *, enum muiReturnValue)){
    muiObject *b;
    b = muiNewButton (x, x+width, y, y+height);
    muiLoadButton (b, name);
    muiSetCallback (b, callback);
    muiSetID (b, id);
    return b;
}
    
void makeButtons (int win){
    muiObject *prevb, *b;
    int x, yh, yl;
    int maxX, maxY;
    int buttonWidth = 90;
    int buttonHeight = 25;
    int buttonSpacingX = 100;
    int buttonSpacingY = 35;
    glutSetWindow (win);
    maxX = glutGet (GLUT_WINDOW_WIDTH);
    maxY = glutGet (GLUT_WINDOW_HEIGHT);
    maxY -= 23;			       /* reserved space for menubar */
    
    x = 10;
    yh = maxY - buttonSpacingY;
    yl = yh - buttonSpacingY;
    b = makeButton (x, yh, buttonWidth, buttonHeight, MENU_FLY, "View (fly)", buttonFunc);
    b = makeButton (x, yl, buttonWidth, buttonHeight, MENU_EDIT_LAND, "Edit land", buttonFunc);
    x += buttonSpacingX;
    b = makeButton (x, yh, buttonWidth, buttonHeight, MENU_INSERT_OBJECT, "Insert object", buttonFunc);
    b = makeButton (x, yl, buttonWidth, buttonHeight, MENU_EDIT_OBJECT, "Edit object", buttonFunc);
    
    x += buttonSpacingX;
    ModeLabel = muiNewBoldLabel (x, yh, "Mode: view(polar)");
    x += 2*buttonSpacingX - 50;
    ExtraButtons[0] = makeButton (x, yh, buttonWidth, buttonHeight, MENU_FLY, "", buttonFunc);
    ExtraButtons[1] = makeButton (x, yl, buttonWidth, buttonHeight, MENU_FLY, "", buttonFunc);
    x += buttonSpacingX;
    ExtraButtons[2] = makeButton (x, yh, buttonWidth, buttonHeight, MENU_FLY, "", buttonFunc);
    ExtraButtons[3] = makeButton (x, yl, buttonWidth, buttonHeight, MENU_FLY, "", buttonFunc);
    showViewButtons();
}
void makeMenu(int win){
    muiObject *menu;
    int menuFile, menuSize, menuNew, menuMotion, menuEdit, menuMode,
      menuLand;
    
    glutSetWindow (win);
    menu = muiNewPulldown ();
    
    menuFile = glutCreateMenu (menuFunc);
    glutAddMenuEntry ("Load", MENU_LOAD);
    glutAddMenuEntry ("Save", MENU_SAVE);
    glutAddMenuEntry ("Quit", MENU_QUIT);
    
    menuSize = glutCreateMenu (menuFunc);
    glutAddMenuEntry ("16x16", MENU_16);
    glutAddMenuEntry ("32x32", MENU_32);
    glutAddMenuEntry ("64x64", MENU_64);
    glutAddMenuEntry ("128x128", MENU_128);
    
    menuNew = glutCreateMenu (menuFunc);
    glutAddMenuEntry ("Flat", MENU_FLAT);
    glutAddMenuEntry ("Flat (fractal)", MENU_FRACT_1);
    glutAddMenuEntry ("Hills (fractal)", MENU_FRACT_2);
    glutAddMenuEntry ("Mountains (fractal)", MENU_FRACT_3);
    
    menuLand = glutCreateMenu (menuFunc);
    glutAddSubMenu ("Size", menuSize);
    glutAddSubMenu ("New land", menuNew);
    
    menuMotion = glutCreateMenu (menuFunc);
    glutAddMenuEntry ("Polar", MENU_POLAR);
    glutAddMenuEntry ("Fly", MENU_FLY);
    glutAddMenuEntry ("Walk", MENU_WALK);

    menuMode = glutCreateMenu (menuFunc);
    glutAddSubMenu ("View", menuMotion);
    glutAddMenuEntry ("Edit land", MENU_EDIT_LAND);
    glutAddMenuEntry ("Edit object", MENU_EDIT_OBJECT);
    glutAddMenuEntry ("Insert object", MENU_INSERT_OBJECT);
    glutAddMenuEntry ("Delete object", MENU_DELETE_OBJECT);
    

    muiAddPulldownEntry (menu, "File", menuFile, 0);
    muiAddPulldownEntry (menu, "Land", menuLand, 0);
#if 0
    muiAddPulldownEntry (menu, "Mode", menuMode, 0);
#endif    
    makeButtons (win);
}
