#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <GL/glut.h>
#include <mui/mui.h>
#include "viz.h"
#include "landviewer.h"
#include "vge.h"

#define BUFSIZE 512

extern void makeMenu(int);
extern muiObject *t1, *t2, *t3;	/* from Kuby */
extern void updateAll ();

int View;
static int Wa, Ha;

static int Selected;
static int LandList;
static int MaterialList;
static int SelectFlagList;
int NumObjs;
TexObj *Objs;

static float AreaX1, AreaX2, AreaZ1, AreaZ2;
static float SelectX, SelectY, SelectZ;
static float PointX, PointZ;
static int OldX, OldY;
static int Raising, SelectingArea, AreaSelected;
static int Dragging = 0;
static int Moving = 0, WasSelected = -1;

void drawLand (GLenum mode){
    if (mode==GL_SELECT){
	glLoadName (1);
    }
    glMatrixMode (GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glTranslatef (0.0, -0.5, 0.0);
    glCallList (LandList);
    glPopMatrix();
}
static float Distance (float x1, float y1, float z1, 
		float x2, float y2, float z2){
    return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
}

void drawObjects (GLenum mode){
    int i;
    int repX, repY, repZ;
    int mx, my, mz;
    int lod;
    float dst;
    float sc;
    for (i=0; i<NumObjs; i++){
	if (mode==GL_SELECT){
	    glLoadName (i);
	}
	glMatrixMode (GL_MODELVIEW);
	glPushMatrix();
	mx = Objs[i].t.rep[0];
	my = Objs[i].t.rep[1];
	mz = Objs[i].t.rep[2];
	if (mx==-1) mx = 10000;
	if (my==-1) my = 10000;
	if (mz==-1) mz = 10000;
	for (repX=0; repX<mx; repX++){
	    for (repY=0; repY<my; repY++){
		for (repZ=0; repZ<mz; repZ++){
		    dst = Distance (Ex, Ey, Ez, 
				    Objs[i].pos[0]+repX * Objs[i].t.bBox[0],
				    Objs[i].pos[1]+repY * Objs[i].t.bBox[1],
				    Objs[i].pos[2]+repZ * Objs[i].t.bBox[2]);
		    if (Objs[i].t.isVRMLLOD){
			if (dst<(float)Objs[i].t.radius[0]/20.0 || Objs[i].t.radius[0]<0){
			    lod = 0;
			} else if (dst<(float)Objs[i].t.radius[1]/20.0 || Objs[i].t.radius[1]<0){
			    lod = 1;
			} else if (dst<(float)Objs[i].t.radius[2]/20.0 || Objs[i].t.radius[2]<0){
			    lod = 2;
			} else {
			    lod = 3;
			}
			sc = 1.0;
		    } else {
			if (dst<0.5) {lod = 0; sc = 1.0;}
			else if (dst<0.9) {lod = 1; sc = 2.0; }
			else if (dst<3.0) {lod = 2; sc = 4.0; }
			else {lod = 3; sc = 8.0;}
		    }
		    glLoadIdentity();
		    glTranslatef (Objs[i].pos[0], Objs[i].pos[1], Objs[i].pos[2]);
		    glRotatef (Objs[i].t.rot[0], 1.0, 0.0, 0.0);
		    glRotatef (Objs[i].t.rot[1], 0.0, 1.0, 0.0);
		    glRotatef (Objs[i].t.rot[2], 0.0, 0.0, 1.0);
		    glTranslatef (repX * Objs[i].t.bBox[0],
				  repY * Objs[i].t.bBox[1],
				  repZ * Objs[i].t.bBox[2]);
		    if ((repZ * Objs[i].t.bBox[2] + Objs[i].pos[2]) > 1.0) break;
		    glScalef (0.005, 0.005, 0.005);
		    glScalef (Objs[i].t.scale[0],Objs[i].t.scale[1],Objs[i].t.scale[2]);
		    glScalef ( sc, sc, sc);   /* lod scale */
		    glCallList (Objs[i].list[lod]);
		}
		if ((repY * Objs[i].t.bBox[1] + Objs[i].pos[1]) > 1.0) break;
	    }
	    if ((repX * Objs[i].t.bBox[0] + Objs[i].pos[0]) > 1.0) break;
	}
	glPopMatrix();
    }
}

void displayView (){
    glClearColor (0.0, 0.0, 0.0, 1.0);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode (GL_PROJECTION);
    glPopMatrix();
    glPushMatrix();
    agvViewTransform();
    glEnable (GL_LIGHTING);
    drawLand (GL_RENDER);
    drawObjects (GL_RENDER);
    glDisable (GL_LIGHTING);
    if (Selected != -1){
	glLoadIdentity();
	if (Selected == -2){
	    glTranslatef (SelectX, SelectY, SelectZ);
	} else {
	    glTranslatef (Objs[Selected].pos[0],
			  Objs[Selected].pos[1],
			  Objs[Selected].pos[2]);
	    glRotatef (Objs[Selected].t.rot[0], 1.0, 0.0, 0.0);
	    glRotatef (Objs[Selected].t.rot[1], 0.0, 1.0, 0.0);
	    glRotatef (Objs[Selected].t.rot[2], 0.0, 0.0, 1.0);
	    glScalef (1.0, 1.0, 1.0); 
	    glTranslatef ((float)Objs[Selected].t.rep[0]/2.0 * Objs[Selected].t.bBox[0],
			  (float)Objs[Selected].t.rep[1]/2.0 * Objs[Selected].t.bBox[1],
			  (float)Objs[Selected].t.rep[2]/2.0 * Objs[Selected].t.bBox[2]);
	}
	glCallList (SelectFlagList);
    }
    if (AreaSelected){
	glLoadIdentity();
	glBegin (GL_LINES);
	glColor3f (1.0, 1.0, 1.0);
	glVertex3f (AreaX1, -0.5, AreaZ1);
	glVertex3f (AreaX1, 0.5, AreaZ1);
	glVertex3f (AreaX1, -0.5, AreaZ2);
	glVertex3f (AreaX1, 0.5, AreaZ2);
	glVertex3f (AreaX2, -0.5, AreaZ2);
	glVertex3f (AreaX2, 0.5, AreaZ2);
	glVertex3f (AreaX2, -0.5, AreaZ1);
	glVertex3f (AreaX2, 0.5, AreaZ1);
	glEnd();
    }
    glutSwapBuffers();
}

void reshapeView (int w, int h){
    glViewport (0,0, w, h);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluPerspective (60.0, (GLdouble)w/h, 0.01, 100);
    glPushMatrix();
    glMatrixMode (GL_MODELVIEW);
    Wa = w; 
    Ha = h;
}
void initView (int win){
    /* material */
    GLfloat mtn_ambuse[] =   { 0.426, 0.256, 0.108, 0.0 };
    GLfloat mtn_specular[] = { 0.394, 0.272, 0.167, 1.0 };
    GLfloat mtn_shininess[] = { 10 };
    int i;
    int j;
    /* lights */
    GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
    GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat lmodel_ambient[] = { 0.3, 0.3, 0.3, 0.0 };
    GLfloat light_position[] = { 2.0, 1.0, 0.0, 0.0 };
    GLfloat light_position1[] = { -1.7, 1.0, 1.0, 0.0 };
    GLfloat light_position2[] = { -1.7, 1.0, -1.0, 0.0 };
    
    glutSetWindow (win);
    View = win;
    MaterialList =  glGenLists (1);
    glNewList(MaterialList, GL_COMPILE);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mtn_ambuse); 
    glMaterialfv(GL_FRONT, GL_SPECULAR, mtn_specular); 
    glMaterialfv(GL_FRONT, GL_SHININESS, mtn_shininess); 
    glEndList();

    glutDisplayFunc (displayView);
    glutReshapeFunc (reshapeView);
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity();
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    
    glLightfv(GL_LIGHT1, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT1, GL_POSITION, light_position1);
    
    glLightfv(GL_LIGHT2, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT2, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT2, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT2, GL_POSITION, light_position2);
    
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
   
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glEnable(GL_LIGHT2); 

    glEnable (GL_DEPTH_TEST);
    glDepthFunc (GL_LEQUAL);
    glEnable (GL_NORMALIZE);
    glDisable (GL_CULL_FACE);
    glShadeModel (GL_SMOOTH);
    glDisable (GL_BLEND);

    LandList = glGenLists (1);
    CreateLandList (LandList, MaterialList);
    /****************/
    NumObjs = 0;
    Objs = NULL;
    Selected = -1;
    SelectFlagList = glGenLists (1);
    glNewList (SelectFlagList, GL_COMPILE);
    glEndList();
}

void landChanged (){
    glutSetWindow (View);
    CreateLandList (LandList, MaterialList);
    glutPostRedisplay ();
}
/*****************************************************/
/************     Operations with Obj   **************/
/*****************************************************/
extern void getTextureSize(char *fn, Resolution *sx, Resolution *sy, Resolution *sz);

void updateObj (){
    int i;
    int lastValidLOD = 0;
    char fName[256];
    int rx, ry, rz;
    if (Selected < 0){
	return;
    }
    glutSetWindow (View);

    for (i=0; i<4; i++) glDeleteLists (Objs[Selected].list[i], 1);
    
    Objs[Selected].t = t;
    
    strcpy (fName, Objs[Selected].t.path);
    strcat (fName, Objs[Selected].t.fName[0]);

    getTextureSize( fName,  &rx, &ry, &rz);

    if (sscanf (muiGetTBString (t1), "%i", Objs[Selected].t.rep) != 1)
      Objs[Selected].t.rep[0] = 1;
    if (sscanf (muiGetTBString (t2), "%i", Objs[Selected].t.rep+2) != 1)
      Objs[Selected].t.rep[2] = 1;
    if (sscanf (muiGetTBString (t3), "%i", Objs[Selected].t.rep+1) != 1)
      Objs[Selected].t.rep[1] = 1;

    Objs[Selected].t.bBox[2] = (float)rx * 0.005 * t.scale[0];
    Objs[Selected].t.bBox[0] = (float)ry * 0.005 * t.scale[1];
    Objs[Selected].t.bBox[1] = (float)rz * 0.005 * t.scale[2];
     for (i=0; i<4; i++){
	/************************************************/
	/************************************************/
	if (!t.isVRMLLOD){
	    strcpy (fName, Objs[Selected].t.path);
	    strcat (fName, Objs[Selected].t.fName[i]);
	    Objs[Selected].list[i] = ConvertVoxelObject2Faces ( fName, i, 0);
	    printf ("normal lod\n");
	} else {
	    strcpy (fName, Objs[Selected].t.path);
	    printf ("%s\n", Objs[Selected].t.fName[i]);
	    if (strcmp (Objs[Selected].t.fName[i], "")){
		strcat (fName, Objs[Selected].t.fName[i]);
		Objs[Selected].list[i] = ConvertVoxelObject2Faces ( fName, 0, 0);
		lastValidLOD = i;
	    } else {
		if (i==0){
		    fprintf (stderr, "error : no LOD 0 defined\n");
		    exit (1);
		}
		strcat (fName, Objs[Selected].t.fName[lastValidLOD]);
		Objs[Selected].list[i] = ConvertVoxelObject2Faces ( fName, 0, 0);
	    }
	    printf ("VRMLLOD\n");
	}
    }
    glutPostRedisplay();
}
void addObject(float x, float y, float z){
    int i;
    int lastValidLOD;
    char fName[256];
    int rx, ry, rz;
    glutSetWindow (View);
    if (Objs==NULL){
	Objs = (TexObj*)malloc (sizeof(TexObj));
	NumObjs = 0;
    } else {
	Objs = (TexObj*)realloc (Objs, sizeof(TexObj) * (NumObjs+1));
    }
    Objs[NumObjs].pos[0] = x;
    Objs[NumObjs].pos[1] = y;
    Objs[NumObjs].pos[2] = z;
    Objs[NumObjs].t = t;
 
    strcpy (fName, Objs[NumObjs].t.path);
    strcat (fName, Objs[NumObjs].t.fName[0]);

    getTextureSize( fName,  &rx, &ry, &rz);
    

    if (sscanf (muiGetTBString (t1), "%i", Objs[NumObjs].t.rep) != 1)
      Objs[NumObjs].t.rep[0] = 1;
    if (sscanf (muiGetTBString (t2), "%i", Objs[NumObjs].t.rep+2) != 1)
      Objs[NumObjs].t.rep[2] = 1;
    if (sscanf (muiGetTBString (t3), "%i", Objs[NumObjs].t.rep+1) != 1)
      Objs[NumObjs].t.rep[1] = 1;

    Objs[NumObjs].t.bBox[2] = (float)rx * 0.005 * t.scale[0];
    Objs[NumObjs].t.bBox[0] = (float)ry * 0.005 * t.scale[1];
    Objs[NumObjs].t.bBox[1] = (float)rz * 0.005 * t.scale[2];
    
    lastValidLOD = 0;
    for (i=0; i<4; i++){
	/************************************************/
	/************************************************/
	if (!t.isVRMLLOD){
	    strcpy (fName, Objs[NumObjs].t.path);
	    strcat (fName, Objs[NumObjs].t.fName[i]);
	    Objs[NumObjs].list[i] = ConvertVoxelObject2Faces ( fName, i, 0);
	    printf ("normal lod\n");
	} else {
	    strcpy (fName, Objs[NumObjs].t.path);
	    printf ("%s\n", Objs[NumObjs].t.fName[i]);
	    if (strcmp (Objs[NumObjs].t.fName[i], "")){
		strcat (fName, Objs[NumObjs].t.fName[i]);
		Objs[NumObjs].list[i] = ConvertVoxelObject2Faces ( fName, 0, 0);
		lastValidLOD = i;
	    } else {
		if (i==0){
		    fprintf (stderr, "error : no LOD 0 defined\n");
		    exit (1);
		}
		strcat (fName, Objs[NumObjs].t.fName[lastValidLOD]);
		Objs[NumObjs].list[i] = ConvertVoxelObject2Faces ( fName, 0, 0);
	    }
	    printf ("VRMLLOD\n");
	}
    }
    NumObjs++;
    glutPostRedisplay();
}
void deleteObject (){
    int i;
    if (Selected < 0) return;
    for (i=Selected; i<NumObjs-1; i++){
	Objs[i] = Objs[i+1];
    }
    NumObjs--;
    Objs = (TexObj*) realloc (Objs, sizeof(TexObj)*NumObjs);
    if ( NumObjs == 0) Objs = NULL;
    Selected = -1;
    glutPostWindowRedisplay (View);
}
void fallObject(){
    int ix, iz;
    int d;

    if (Selected!=-1){
	d = Dim (Level);
	ix = (int) (((Objs[Selected].pos[0]+1.0)/2.0)*(float)d);
	iz = (int) (((Objs[Selected].pos[1]+1.0)/2.0)*(float)d);
	Objs[Selected].pos[1] = Height[IndexOf (ix,iz, Level)][1] - 0.5;
	SelectY = Objs[Selected].pos[1];
	glutPostWindowRedisplay (View);
    }
}
void raiseObject(){
    if (Selected!=-1){
	RaiseRegionTo (Objs[Selected].pos[0], Objs[Selected].pos[2],
		       Objs[Selected].pos[0] + Objs[Selected].t.bBox[0]*(float)Objs[Selected].t.rep[0],
		       Objs[Selected].pos[2] + Objs[Selected].t.bBox[2]*(float)Objs[Selected].t.rep[2],
		       Objs[Selected].pos[1]);
	glutPostWindowRedisplay (View);
    }
}

void createSelectList (int o){
    if (o<0){
	glNewList (SelectFlagList, GL_COMPILE);
	glColor3f (1.0, 1.0, 1.0);
	glTranslatef (0.0, 0.1, 0.0);
	glScalef (0.1, 0.2, 0.1);
	glutWireCube (1.0);
	glEndList();
    } else {
	glNewList (SelectFlagList , GL_COMPILE);
	glColor3f (1.0, 1.0, 1.0);
	glScalef (Objs[o].t.bBox[0] * Objs[o].t.rep[0],
		  Objs[o].t.bBox[1] * Objs[o].t.rep[1],
		  Objs[o].t.bBox[2] * Objs[o].t.rep[2]);
	glutWireCube (1.0);
	glEndList();
    }
}

void editMouse (int button, int state, int x, int y){
    GLuint selectBuf[BUFSIZE];
    GLint hits;
    GLint viewport[4];
    GLdouble ox, oy, oz, mm[16], pm[16], wx, wy, wz;
    GLuint *ptr;
    
    if (button != GLUT_LEFT_BUTTON || state != GLUT_DOWN)
      return;
    
    glGetIntegerv(GL_VIEWPORT, viewport);
    glSelectBuffer(BUFSIZE, selectBuf);
    (void) glRenderMode(GL_SELECT);
    glInitNames();
    glPushName((GLuint) ~0);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y),
		  5.0, 5.0, viewport);
    gluPerspective(60, (GLdouble)Wa/Ha, 0.01, 100);
    agvViewTransform();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    drawObjects (GL_SELECT); 
    glFlush();
    hits = glRenderMode(GL_RENDER);
    if (hits==0) {
	return;
    }
    glGetDoublev (GL_PROJECTION_MATRIX, pm);
    glGetDoublev (GL_MODELVIEW_MATRIX, mm);
    ptr = (GLuint *) selectBuf;
    ptr++;
    wz = (float) *ptr /0xffffffff;
    ptr+=2;			       /* skip number, z1, z2 */
    Selected = *ptr;
    createSelectList (Selected);
    t = Objs[Selected].t; 
    updateAll(); 		       /* update object prop window */
    glutPostWindowRedisplay(View);
}

int landIntersection (int x, int y, GLdouble *ox, GLdouble *oy, GLdouble *oz){
    GLuint selectBuf[BUFSIZE];
    GLint hits;
    GLint viewport[4];
    GLdouble mm[16], pm[16], wx, wy, wz;
    GLuint *ptr;
    
    glGetIntegerv(GL_VIEWPORT, viewport);
    glSelectBuffer(BUFSIZE, selectBuf);
    (void) glRenderMode(GL_SELECT);
    glInitNames();
    glPushName((GLuint) ~0);
    /*------------------------------------*/
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y),
		  5.0, 5.0, viewport);
    gluPerspective(60, (GLdouble)Wa/Ha, 0.01, 100);
    agvViewTransform();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    drawLand(GL_SELECT);
    glFlush();
    hits = glRenderMode(GL_RENDER);
    if (hits==0) {
	return 0;
    }
    glGetDoublev (GL_PROJECTION_MATRIX, pm);
    glGetDoublev (GL_MODELVIEW_MATRIX, mm);
    ptr = (GLuint *) selectBuf;
    ptr++;
    wx = x;
    wy = y;
    wz = (float) *ptr /0xffffffff;
    ptr+=2;			       /* skip number, z1, z2 */
    gluUnProject(wx, wy, wz, mm, pm, viewport, ox, oy, oz);
    return 1;
}

void selectLand (int button, int state, int x, int y){
    GLdouble ox, oy, oz;
    
    if (button == GLUT_RIGHT_BUTTON)
      return;
    if (state == GLUT_UP){
	Raising = 0;
	if (SelectingArea){
	    SelectingArea = 0;
	    if (fabs (AreaX1-AreaX2) < 0.05 ||
		fabs (AreaZ1-AreaZ2) < 0.05)
	      AreaSelected = 0;
	    glutSetWindow (View);
	    glutPostRedisplay();
	}
	return;
    }
    if (!landIntersection (x, y, &ox, &oy, &oz)){
	return;
    }
    
    glutPostRedisplay();
    OldX = x;
    OldY = y;
    PointX = ox;
    PointZ = oz;
    if (button == GLUT_LEFT_BUTTON &&
	state == GLUT_DOWN) Raising = 1;
    if (button==GLUT_MIDDLE_BUTTON &&
	state == GLUT_DOWN){
	SelectingArea = 1;
	AreaSelected = 1;
	AreaX1 = AreaX2 = ox;
	AreaZ1 = AreaZ2 = oz;
    }
}
void levelRegion(){
    if (AreaSelected)
      RaiseRegionTo (AreaX1, AreaZ1, AreaX2, AreaZ2, 0.0);
}
void selectLandDrag (int x, int y){
    if (Raising){
	if (!AreaSelected){
	    Raise (PointX, PointZ, -(y-OldY));
	} else {
	    RaiseRegion (AreaX1, AreaZ1, AreaX2, AreaZ2, OldY-y);
	}
	OldY = y;
	glutPostWindowRedisplay (View);
	return;
    }
    if (SelectingArea){
	GLdouble ox, oy, oz;
	if (!landIntersection (x, y, &ox, &oy, &oz)){
	    return;
	}
	AreaX2 = ox;
	AreaZ2 = oz;
	if (fabs(AreaX2)>1.0) AreaX2 = AreaX2 / fabs(AreaX2);   /* =1.0 keeping sign */
	if (fabs(AreaZ2)>1.0) AreaZ2 = AreaZ2 / fabs(AreaZ2);   /* =1.0 keeping sign */
	OldY = y;
	OldX = x;
	glutPostRedisplay();
    }
}
void insertPassiveMotion (int x, int y){
    GLdouble ox, oy, oz;
    Selected = -2;
    if (!landIntersection (x, y, &ox, &oy, &oz)){
	Selected = -1;
    } else {
	SelectX = ox;
	SelectY = oy;
	SelectZ = oz;
    }
    glutPostRedisplay();
}
void insertMouse (int button, int state, int x, int y){
    if (button != GLUT_LEFT_BUTTON){
	return;
    }
    if (state == GLUT_DOWN){
	Dragging = 1;
	OldY = y;
	return;
    }
    if (state == GLUT_UP){
	Dragging = 0;
	if (Moving){
	    Moving = 0;
	    Objs[WasSelected].pos[0] = SelectX - (float)Objs[WasSelected].t.rep[0]/2.0 * Objs[WasSelected].t.bBox[0]; 
	    Objs[WasSelected].pos[1] = SelectY - (float)Objs[WasSelected].t.rep[1]/2.0 * Objs[WasSelected].t.bBox[1];
	    Objs[WasSelected].pos[2] = SelectZ - (float)Objs[WasSelected].t.rep[2]/2.0 * Objs[WasSelected].t.bBox[2];
	    printf ("endMoving Selected = %i\n", WasSelected);
	    Selected = WasSelected;
	    glutPostWindowRedisplay (View);
	    menuFunc (MENU_EDIT_OBJECT);
	} else {
	    addObject (SelectX, SelectY, SelectZ);
	    menuFunc (MENU_FLY);
	}
    }
}
void insertMotion (int x, int y){
    if (Dragging){
	SelectY += (OldY - y) * 0.01;
	OldY = y;
	glutPostRedisplay();
    }
}
void setMode (int m){
    glutSetWindow (View);
    switch (m){
     case MENU_EDIT_LAND:
	glutMotionFunc (selectLandDrag);
	glutPassiveMotionFunc (NULL);
	glutMouseFunc (selectLand);
	Selected = -1;
	break;
     case MENU_POLAR:
	glutMotionFunc (agvHandleMotion);
	glutPassiveMotionFunc (NULL);
	glutMouseFunc (agvHandleButton);
	agvSwitchMoveMode (POLAR);
	Selected = -1;
	break;
     case MENU_WALK:
 	agvSetHeightFunc (heightAt);
	agvSwitchMoveMode (WALKING);
	glutPassiveMotionFunc (NULL);
	glutMotionFunc (agvHandleMotion);
	glutMouseFunc (agvHandleButton);
	Selected = -1;
	break;
     case MENU_FLY:
	agvSwitchMoveMode (FLYING);
	glutPassiveMotionFunc (NULL);
	glutMotionFunc (agvHandleMotion);
	glutMouseFunc (agvHandleButton);
	Selected = -1;
	break;
     case MENU_INSERT_OBJECT:
	createSelectList(-1);
	glutPassiveMotionFunc (insertPassiveMotion);
	glutMotionFunc (insertMotion);
	glutMouseFunc (insertMouse);
	break;
     case MENU_EDIT_OBJECT:
	glutPassiveMotionFunc (NULL);
	glutMotionFunc (NULL);
	glutMouseFunc (editMouse);
	break;
     case MENU_MOVE_OBJECT:
	if (Selected >= 0){
	    WasSelected = Selected;
	    glutPassiveMotionFunc (insertPassiveMotion);
	    glutMotionFunc (insertMotion);
	    glutMouseFunc (insertMouse);
	    Moving = 1;
	}
	break;
	
   }
    glutPostRedisplay();
}
