/***************************************************************************/
/*                                                                         */
/*                b3dt - 3D texture file builder v0.23                     */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*  txengine.h - header file for txengine.c (contain main texture engine)  */
/*                                                                         */
/***************************************************************************/

#ifndef _TXENGINE_H_
#define _TXENGINE_H_ 1

/* some miscellaneous constants */

/* logical status */
#define NO 0
#define YES 1

/* define thickness of object walls */
#define THICKNESS 2 

/* symbolic constants for object types */
#define	HIGH_HOUSE 0
#define	BUILDING_HOUSE 1
#define	BOX_HOUSE 2
#define	COUNTRY_HOUSE 3
#define	WATER_WORKS 4
#define TREE 5
#define BUSH 6

/* output texture file name buffer */
#define FILENAME_SIZE 256

/* size of texture type file header */
#define SIZE_OF_FILE_HEADER 4

/* standard texture file header */
#define STANDARD_HEADER "# 3d texture file v1.0"
#define OBJECT_HEADER "# object index type: " 

/* symbolic constants for voxel colors */
#define EMPTY_COLOR 0
#define WALL_COLOR 1
#define WINDOW_COLOR 2
#define ROOF_COLOR 3
#define TREE_STEM_COLOR 4
#define TREE_TOP_COLOR 5
#define BUSH_COLOR 6
/* Get color of voxel */
int GetVoxelColor(int partOfObject);

/* Write color palette to texture file */
void WritePalette(int tObject);


/* set up texture type, may be: vtf, stf or ctf */
void SetTexture(int selectedTextureType);

/* set up texture type, may be: highhouse */
/* buildinghouse, boxhouse, countryhouse  */
/* or waterworks                          */
void SetObject(int selectedObjectType);

/* set output texture file name */
void SetFileName(char *selectedFileName);

/* function which generate voxels for */
/* high house object type              */
void GenerateHighHouse(void);

/* function which generate voxels for */
/* building house object type         */
void GenerateBuildingHouse(void);

/* function which generate voxels for */
/* building "box" house object type   */
void GenerateBoxHouse(void);

/* function which generate voxels for */
/* country house object type   */
void GenerateCountryHouse(void);

/* function which generate voxels for */
/* water works object type            */
void GenerateWaterWorks(void);

/* function which generate voxels for */
/* tree object type                   */
void GenerateTree(void);

/* function which generate voxels for */
/* bush object type                   */
void GenerateBush(void);

/* main texture function, which call */
/* one function on top               */
void Build3DTexture(void);

#endif
/* End of file: object.h */
