/***************************************************************************/
/*                                                                         */
/*                b3dt - 3D texture file builder v0.23                     */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*       contrib.c - module to b3dt (contain auxiliary functions)          */
/*                                                                         */
/***************************************************************************/

/* include header files */

/* standard system header files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* user defined header files */
#include "contrib.h"
#include "txengine.h"

/* include header files specific for unix only */
#if defined(UNIX)
#include <unistd.h>
#include <sys/utsname.h>
#endif

/* global variables */

/* command line args for help */
const char *strHelp = "-help";
const char *strLongHelp = "--help";

/* command line args for texture types */
const char *strVTF = "-vtf";
const char *strSTF = "-stf";
const char *strCTF = "-ctf";

/* command line args for object types */
const char *strHighHouse = "-highhouse";
const char *strBuildingHouse = "-buildinghouse";
const char *strBoxHouse = "-boxhouse";
const char *strCountryHouse = "-countryhouse";
const char *strWaterWorks = "-waterworks";
const char *strTree = "-tree";
const char *strBush = "-bush";

/* flag, which detect if texture type is selected */
int isTexture = YES;

/* flag, which detect if object type is selected */
int isObject = YES;

/* function for putting message on terminal */
void PutMessage(int controlPoint){

	switch (controlPoint){
		case B3DT_START: /* say, that we are starting */
				 printf("%s", MSG_START);
				 break;
				 
		case B3DT_END:   /* say, that we are exiting */
				 printf("%s", MSG_END);
				 break;
				 
		case B3DT_USAGE: /* say, how to use this program */ 
				 printf("%s", MSG_USAGE);
				 printf("\n");
				 exit(OK);
				 break;
				 
		case B3DT_FILE_ERR: /* say, that we have problems with */
		                 /* opening  the output  texture  file */ 
				 printf("%s", MSG_FILE_ERR);
				 break;
				 
		default:         /* say, that we have a critical error */ 
				 printf("%s", MSG_ERR);
	}
}

/* function for setting up unix environment */
void InitEnvironment(void){

	time_t t;
	
	srand((unsigned) time(&t));

#if defined(UNIX)
	/* get miscellaneous unix environment infos */
	/* that is process PID, PPID, user UID, GID */
	printf("\n%s%d%s%d%c", "process id (", getpid(), "), process parent id (", getppid(), ')');
	printf("\n%s%d%s%d%c", "user uid (", getuid(), "), user gid (", getgid(), ')');
#endif

}

/* function for putting help infos on terminal */
void GetHelp(void){

	printf("%s", MSG_USAGE);
	printf("\n%s", "[object type] is:  -highhouse");
	printf("\n\t\t  %s", " -buildinghouse");
	printf("\n\t\t  %s", " -boxhouse");
	printf("\n\t\t  %s", " -countryhouse");
	printf("\n\t\t  %s", " -waterworks");
	printf("\n\t\t  %s", " -tree");
	printf("\n\t\t  %s", " -bush");
	printf("\n%s", "[texture type] is: -vtf (voxel texture format)");
	printf("\n\t\t  %s", " -stf (surface texture format)");
	printf("\n\t\t  %s\n", " -ctf (column texture format)");
	exit(OK);
}

/* function for parsing command line arguments */
void ParseArg(int index, char *arg){

	/* test, if user want a help message */
	if (!strcmp(arg,strHelp) || !strcmp(arg,strLongHelp)){
		GetHelp();
		exit(OK);
	}

	switch (index){
		/* parse for texture type */
		case TEXTURE_TYPE: isTexture = NO;
						   if (!strcmp(arg,strVTF)){
							   SetTexture(VTF_TEXTURE);
							   isTexture = YES;
						   }
						   if (!strcmp(arg,strSTF)){
							   SetTexture(STF_TEXTURE);
							   isTexture = YES;
						   }
						   if (!strcmp(arg,strCTF)){
							   SetTexture(CTF_TEXTURE);
							   isTexture = YES;
						   }
						   break;

		/* parse for object type */						   
		case OBJECT_TYPE:  isObject = NO;
						   if (!strcmp(arg,strHighHouse)){
							   SetObject(HIGH_HOUSE);
							   isObject = YES;
						   }
						   if (!strcmp(arg,strBuildingHouse)){
							   SetObject(BUILDING_HOUSE);
							   isObject = YES;
						   }
						   if (!strcmp(arg,strBoxHouse)){
							   SetObject(BOX_HOUSE);
							   isObject = YES;
						   }
						   if (!strcmp(arg,strCountryHouse)){
							   SetObject(COUNTRY_HOUSE);
							   isObject = YES;
						   }
						   if (!strcmp(arg,strWaterWorks)){
							   SetObject(WATER_WORKS);
							   isObject = YES;
						   }
						   if (!strcmp(arg,strTree)){
						   	   SetObject(TREE);
						   	   isObject = YES;
						   }
						   if (!strcmp(arg,strBush)){
						   	   SetObject(BUSH);
						   	   isObject = YES;
						   }	    	   
						   break;
						   
		/* parse for output file name */				   
		case OUTPUT_FILE:  SetFileName(arg);
						   break;
		default: return;
	}

	/* if unknown texture on command line, than exit */
	if (!isTexture){
		printf("\n%s%s\n","Unknown texture format: ", arg);
		exit(ERROR);
	}

	/* if unknown object on command line, than exit too */
	if (!isObject){
		printf("\n%s%s\n","Unknown object type: ", arg);
		exit(ERROR);
	}
}

/* End of file: contrib.c */
