/***************************************************************************/
/*                                                                         */
/*                b3dt - 3D texture file builder v0.23                     */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*         txengine.c - module to b3dt (contain main texture engine)       */
/*                                                                         */
/***************************************************************************/

/* include header files */

/* standard system header files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* user defined header files */
#include "contrib.h"
#include "txengine.h"

/* include header files specific for unix only */
#if defined(UNIX)
#include <sys/stat.h>
#include <unistd.h>
#endif               

/* global variables */

/* output texture file name */
char fileName[FILENAME_SIZE];

/* standard texture file header */
char fileHeader[SIZE_OF_FILE_HEADER];

/* variable for object type, set default */
int objectType = HIGH_HOUSE;

/* variable for texture type, set default */
int textureType = VTF_TEXTURE;

/* voxel color components */
int red = 0;
int green = 0;
int blue = 0;

/* string header type on texture file */
char *strFileHeaderVTF = "VTF";
char *strFileHeaderSTF = "STF";
char *strFileHeaderCTF = "CTF";

/* variable for voxel color, set default */
int voxelColor = EMPTY_COLOR;

/* output texture file name handler */
FILE *outputFile;

/* string for visual progress on terminal */
const char inProgress[6] = "|/-\\";

/* set up texture type, may be: vtf, stf or ctf */
void SetTexture(int selectedTextureType){
	textureType = selectedTextureType;
#if defined(DEBUG)
	printf("\n%d", textureType);
#endif
}

/* Get color of voxel */
int GetVoxelColor(int partOfObject){

	int newColor;
	
	newColor = EMPTY_COLOR;
	switch(partOfObject){
		case WALL_COLOR:
			newColor = 1 + rand() % 3;
			break;
		case WINDOW_COLOR:
			newColor = 4 + rand() % 3;
			break;
		case ROOF_COLOR:
			newColor = 7 + rand() % 3;
			break;	
		case TREE_STEM_COLOR:
			newColor = 1 + rand() % 3;
			break;
		case TREE_TOP_COLOR:
			newColor = 4 + rand() % 3;
			break;
		case BUSH_COLOR:
			newColor = 1 + rand() % 3;
			break;		
		default:
			newColor = EMPTY_COLOR;
			break;
	}
	return(newColor);
}		

/* Write color palette to texture file */
void WritePalette(int tObject){

	switch(tObject){

		case HIGH_HOUSE:
			/* palette for wall color */
			fprintf(outputFile, "c %d %d %d\n",  84,  74,  84);
			fprintf(outputFile, "c %d %d %d\n", 114, 114, 114);
			fprintf(outputFile, "c %d %d %d\n", 144, 144, 144);
			/* palette for window color */
			fprintf(outputFile, "c %d %d %d\n", 144, 152, 235);
			fprintf(outputFile, "c %d %d %d\n", 115, 123, 231);
			fprintf(outputFile, "c %d %d %d\n", 206, 206, 247);
			break;

		case BUILDING_HOUSE:
			/* palette for wall color */
			fprintf(outputFile, "c %d %d %d\n",  80,  70,  80);
			fprintf(outputFile, "c %d %d %d\n", 120, 120, 120);
			fprintf(outputFile, "c %d %d %d\n", 150, 150, 150);
			/* palette for window color */
			fprintf(outputFile, "c %d %d %d\n", 144, 152, 235);
			fprintf(outputFile, "c %d %d %d\n", 115, 123, 231);
			fprintf(outputFile, "c %d %d %d\n", 206, 206, 247);
			break;

		case BOX_HOUSE:
			/* palette for wall color */
			fprintf(outputFile, "c %d %d %d\n",  90,  80,  90);
			fprintf(outputFile, "c %d %d %d\n", 110, 110, 110);
			fprintf(outputFile, "c %d %d %d\n", 145, 135, 145);
			/* palette for window color */
			fprintf(outputFile, "c %d %d %d\n", 144, 152, 235);
			fprintf(outputFile, "c %d %d %d\n", 115, 123, 231);
			fprintf(outputFile, "c %d %d %d\n", 206, 206, 247);
			break;					

		case COUNTRY_HOUSE:
			/* palette for wall color */
			fprintf(outputFile, "c %d %d %d\n",  84,  84,  84);
			fprintf(outputFile, "c %d %d %d\n", 114, 114, 114);
			fprintf(outputFile, "c %d %d %d\n", 144, 144, 144);
			/* palette for window color */
			fprintf(outputFile, "c %d %d %d\n", 144, 152, 235);
			fprintf(outputFile, "c %d %d %d\n", 115, 123, 231);
			fprintf(outputFile, "c %d %d %d\n", 206, 206, 247);
			/* palette for roof color */
			fprintf(outputFile, "c %d %d %d\n", 255,   0,   0);
			fprintf(outputFile, "c %d %d %d\n", 200,   5,   5);
			fprintf(outputFile, "c %d %d %d\n", 255,  70,  70);
			break;			

		case WATER_WORKS:
			/* palette for wall color */
			fprintf(outputFile, "c %d %d %d\n",  75,  85,  75);
			fprintf(outputFile, "c %d %d %d\n", 115, 125, 115);
			fprintf(outputFile, "c %d %d %d\n", 140, 150, 140);
			break;

		case TREE:
			/* palette for tree stem color */
			fprintf(outputFile, "c %d %d %d\n", 135, 120, 85);
			fprintf(outputFile, "c %d %d %d\n", 190, 140, 90);
			fprintf(outputFile, "c %d %d %d\n", 175, 120, 75);
			/* palette for tree top color */
			fprintf(outputFile, "c %d %d %d\n",  10, 170, 10);
			fprintf(outputFile, "c %d %d %d\n", 150, 155, 30);
			fprintf(outputFile, "c %d %d %d\n", 105, 160, 10);
			break;

		case BUSH:
			/* palette for bush color */
			fprintf(outputFile, "c %d %d %d\n",  10, 170, 10);
			fprintf(outputFile, "c %d %d %d\n", 145, 160, 25);
			fprintf(outputFile, "c %d %d %d\n",  90, 150,  5);
			break;

		default: 
			break;				
		}
}

/* set up texture type, may be: highhouse */
/* buildinghouse, boxhouse, countryhouse  */
/* or waterworks                          */
void SetObject(int selectedObjectType){

	objectType = selectedObjectType;
#if defined(DEBUG)
	printf("\n%d", objectType);
#endif
}

/* set output texture file name */
void SetFileName(char *selectedFileName){

	strcpy(fileName, selectedFileName);
#if defined(DEBUG)
	printf("\n%s", fileName);
#endif
}

/* function which generate voxels for */
/* high house object type              */ 
void GenerateHighHouse(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 25;
	int maxY = 25;
	int maxZ = 50;
	int status = -1;
	
	printf("\n%s", "object type: high house");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = GetVoxelColor(WALL_COLOR);
				if (z < 45) {

					if ((x == 0) || (x == 24)){
 						if (((y > 4) && (y < 10)) || ((y > 14) && (y < 20))){
							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20)) || ((z > 24) && (z < 30)) || ((z > 34) && (z < 40))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((y == 0) || (y == 24)){
						if (((x > 4) && (x < 10)) || ((x > 14) && (x < 20))){
							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20)) || ((z > 24) && (z < 30)) || ((z > 34) && (z < 40))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}	

					if ((x > THICKNESS) && (x < (maxX - THICKNESS)) && (y > THICKNESS) && (y < (maxY - THICKNESS)) && (z > THICKNESS) && (z < (44 - THICKNESS)))
						voxelColor = EMPTY_COLOR; 
				}
				else{

					if (((x > 9) && (x < 15)) && ((y > 9) && (y < 15)))
						voxelColor = GetVoxelColor(WALL_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}			
}

/* function which generate voxels for */
/* building house object type         */
void GenerateBuildingHouse(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 45;
	int maxY = 25;
	int maxZ = 30;
	int status = -1;

	printf("\n%s", "object type: building house");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = GetVoxelColor(WALL_COLOR);
				if (z < 25) {

					if ((x == 0) || (x == 44)){

						if (((y > 4) && (y < 10)) || ((y > 14) && (y < 20))){

							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((y == 0) || (y == 24)){

						if (((x > 4) && (x < 10)) || ((x > 14) && (x < 20)) || ((x > 24) && (x < 30)) || ((x > 34) && (x < 40))){

							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((x > THICKNESS) && (x < (maxX - THICKNESS)) && (y > THICKNESS) && (y < (maxY - THICKNESS)) && (z > THICKNESS) && (z < (22 - THICKNESS)))
						voxelColor = EMPTY_COLOR; 
				}
				else{

					if ((((x > 9) && (x < 15)) || ((x > 29) && (x < 35))) && ((y > 9) && (y < 15)))
						voxelColor = GetVoxelColor(WALL_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}		
}

/* function which generate voxels for */
/* building "box" house object type   */
void GenerateBoxHouse(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 25;
	int maxY = 25;
	int maxZ = 30;
	int status = -1;

	printf("\n%s", "object type: buildibg \"box\" house");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = GetVoxelColor(WALL_COLOR);
				if (z < 25) {

					if ((x == 0) || (x == 24)){

						if (((y > 4) && (y < 10)) || ((y > 14) && (y < 20))){

							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((y == 0) || (y == 24)){

						if (((x > 4) && (x < 10)) || ((x > 14) && (x < 20))){

							if (((z > 4) && (z < 10)) || ((z > 14) && (z < 20))){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((x > THICKNESS) && (x < (maxX - THICKNESS)) && (y > THICKNESS) && (y < (maxY - THICKNESS)) && (z > THICKNESS) && (z < (22 - THICKNESS)))
						voxelColor = EMPTY_COLOR; 					
				}
				else{

					if (((x > 4) && (x < 10)) && ((y > 14) && (y < 20)))
						voxelColor = GetVoxelColor(WALL_COLOR);
					else
						voxelColor = EMPTY_COLOR;
				}

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}
}

/* function which generate voxels for */
/* country house object type          */
void GenerateCountryHouse(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 25;
	int maxY = 15;
	int maxZ = 25;
	int status = -1;

	printf("\n%s", "object type: country house");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = GetVoxelColor(WALL_COLOR);

				if (z < 15) {

					if ((x == 0) || (x == 24)){

						if ((y > 4) && (y < 10)){

							if ((z > 4) && (z < 10)){

								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((y == 0) || (y == 14)){

						if (((x > 4) && (x < 10)) || ((x > 14) && (x < 20))){

							if ((z > 4) && (z < 10)){
								voxelColor = GetVoxelColor(WINDOW_COLOR);
							}
						}
					}

					if ((x > THICKNESS) && (x < (maxX - THICKNESS)) && (y > THICKNESS) && (y < (maxY - THICKNESS)) && (z > THICKNESS) && (z < (12 - THICKNESS)))

						voxelColor = EMPTY_COLOR; 					
				}
				else{

					if (y < 7){

						if (((10.0/7.0)*((double)(y)) - ((double)(z)) + 15) >= 0)

							voxelColor = GetVoxelColor(ROOF_COLOR);
						else
							voxelColor = EMPTY_COLOR;	 
					}

					else{

						if (((-10.0/7.0)*((double)(y)) - ((double)(z)) + 35) >= 0)
							voxelColor = GetVoxelColor(ROOF_COLOR);

						else
							voxelColor = EMPTY_COLOR;	
					}
				}

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}
}

/* function which generate voxels for */
/* water works object type            */
void GenerateWaterWorks(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 25;
	int maxY = 25;
	int maxZ = 50;
	int status = 0;

	printf("\n%s", "object type: waterworks");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = EMPTY_COLOR;

				if (z < 40) {

					if (((int)pow((double)(x - 12.5),2.0)) + ((int)pow((double)(y - 12.5),2.0)) <= 64)
						voxelColor = GetVoxelColor(WALL_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}
				else{

					if (((int)pow((double)(x - 12.5),2.0)) + ((int)pow((double)(y - 12.5),2.0)) <= 100)
						voxelColor = GetVoxelColor(WALL_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}

				if ((((int)pow((double)(x - 12.5),2.0)) + ((int)pow((double)(y - 12.5),2.0)) <= ((7 - THICKNESS)*(7 - THICKNESS))) && ((z > THICKNESS) && (z < (maxZ - THICKNESS))))
					voxelColor = EMPTY_COLOR;

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}		
}

/* function which generate voxels for */
/* tree object type                   */
void GenerateTree(void){

	int x = 0;
	int y = 0;
	int z = 0;
	int maxX = 10;
	int maxY = 10;
	int maxZ = 15;
	int status = 0;

	printf("\n%s", "object type: tree");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){

		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = EMPTY_COLOR;

				if (z < 8) {

					if (((int)pow((double)(x - 5.0),2.0)) + ((int)pow((double)(y - 5.0),2.0)) <= 3)
						voxelColor = GetVoxelColor(TREE_STEM_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}
				else{
					if (((int)pow((double)(x - 5.0),2.0) + (int)pow((double)(y - 5.0),2.0) + (int)pow((double)(z - 10.5),2.0)) <= (int)(12))
						voxelColor = GetVoxelColor(TREE_TOP_COLOR);

					else
						voxelColor = EMPTY_COLOR;
				}

				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}		
}

/* function which generate voxels for */
/* bush object type                   */
void GenerateBush(void){

	int x = 0;
	int y = 0;
	int z = 0;
	double a = 2.5;
	double b = 2.5;
	double c = 5.0; 
	int maxX = 5;
	int maxY = 5;
	int maxZ = 10;
	int status = 0;

	printf("\n%s", "object type: tree");
	printf("\n%c", '|');

	for(z = 0; z < maxZ; z++){
		status++;
		if (status > 3) status = -1;
		printf("\b%c", inProgress[status]);
		fflush(stdout);

		for(x = 0; x < maxX; x++){

			for(y = 0; y < maxY; y++){

				voxelColor = EMPTY_COLOR;
				if (z <= 1){
					if ((x == 2) && (y == 2))
						voxelColor = GetVoxelColor(BUSH_COLOR);
				}
				else{
					if (((pow((double)x - a,2.0)/a) + (pow((double)y - b,2.0)/b) + (pow((double)z - c,2.0)/7.0))  <= (int)(2.0)) 
							voxelColor = GetVoxelColor(BUSH_COLOR);
						else
							voxelColor = EMPTY_COLOR;
				}
				fprintf(outputFile, "%d ", voxelColor);
			}
		}
	}		

}

/* main texture function, which call */
/* one function on top               */
void Build3DTexture(void){

	/* open output texture file */
	printf("\n%s%s", "opening output texture file: ", fileName);

	/* if some problem, say what's wrong and exit */ 
	if ((outputFile = fopen(fileName, "w+")) == NULL){
		PutMessage(B3DT_FILE_ERR);
		exit(ERROR);
	}

	/* say, that there isn't problems with opening */
	printf("\n%s%s%s", "file ", fileName, " successfully opened");

	/* say, which texture format we are going to write */
	printf("\n%s", "texture format: ");

	switch (textureType){

		case VTF_TEXTURE: printf("%s", "vtf (voxel texture format)");
						  strcpy(fileHeader, strFileHeaderVTF);
						  break;

		case STF_TEXTURE: printf("%s", "stf (surface texture format)");
						  strcpy(fileHeader, strFileHeaderSTF);
						  break;

		case CTF_TEXTURE: printf("%s", "ctf (column texture format)");
						  strcpy(fileHeader, strFileHeaderCTF);
						  break;

		default: break;
	}

	/* write texture file header */
	fprintf(outputFile, "%s %s\n", STANDARD_HEADER, fileHeader);

	/* write voxel data for object, that was selected */
	switch (objectType){
 
 		case HIGH_HOUSE: 
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "1");
				WritePalette(HIGH_HOUSE);
				GenerateHighHouse();
				break;

		case BUILDING_HOUSE:
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "2");
				WritePalette(BUILDING_HOUSE); 
				GenerateBuildingHouse();
				break;

		case BOX_HOUSE: 
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "3");
				WritePalette(BOX_HOUSE);
				GenerateBoxHouse();
				break;

		case COUNTRY_HOUSE:
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "4");
				WritePalette(COUNTRY_HOUSE); 
				GenerateCountryHouse();
				break;

		case WATER_WORKS:
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "5");
				WritePalette(WATER_WORKS); 
				GenerateWaterWorks();
				break;

		case TREE:
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "6");
				WritePalette(TREE); 
				GenerateTree();
				break;

		case BUSH:
				fprintf(outputFile, "%s%s\n", OBJECT_HEADER, "7");
				WritePalette(BUSH); 
				GenerateBush();
				break;

		default: break;
	}
	
	/* say, that we are closeing output texture file */
	printf("\b%s%s", "closeing output texture file ", fileName);
	
	/* if the stream was being used for output, any buffered data */
	/* is  written first, using fflush                            */
	fflush(outputFile);
	
	/* close file */
	if (fclose(outputFile) != OK)
		/* if we have some problem with closeing file */ 
		printf("\n%s", "cannot close output file");

	else	
		printf("\n%s%s%s", "file ", fileName, " closed");
}

/* End of file: object.c */
