/***************************************************************************/
/*                                                                         */
/*                m3dt - 3D texture file mapper v0.23                      */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*  mpengine.h - header file for mpengine.c (contain main texture engine)  */
/*                                                                         */
/***************************************************************************/

#ifndef _MPENGINE_H_
#define _MPENGINE_H_ 1

/* some miscellaneous constants */

/* logical status */
#define NO 0
#define YES 1

/* output texture file name buffer */
#define FILENAME_SIZE 256

/* size of texture type file header */
#define SIZE_OF_FILE_HEADER 4

/* container object description file header */
#define CNT_HEADER "# description scene object file\n"

/* container object description header offset */
#define CNT_OFFSET 32     

/* standard texture file header */
#define STANDARD_HEADER "# 3d texture file v1.0"

/* normalized landscape data file */
#define NORMALIZED_LANDSCAPE_FILE "landscape.ok"

/* texture header offset */
#define HEADER_OFFSET 50

/* maximum of object in scene */
#define MAX_OBJECTS 32

/* horizontal placement */
#define HORIZONTAL_PLACEMENT 0

/* vertical placement */
#define VERTICAL_PLACEMENT 1

/* message size */
#define MSG_SIZE 12
#define MSG_OBJ_SIZE 16

/* boundary around object */
#define BOUNDARY_SIZE 5

/* object types */
#define HIGHHOUSE 1
#define BUILDINGHOUSE 2
#define BOXHOUSE 3
#define COUNTRYHOUSE 4
#define WATERWORKS 5
#define TREE 6
#define BUSH 7

/* geomeric data of objects */

/* HH...high house */
#define HH_SIZE_X 25
#define HH_SIZE_Y 25
#define HH_SIZE_Z 50

/* BH...buildinghouse */
#define BH_SIZE_X 45
#define BH_SIZE_Y 25
#define BH_SIZE_Z 30

/* XH...box house */
#define XH_SIZE_X 25
#define XH_SIZE_Y 25
#define XH_SIZE_Z 30

/* CH...country house */
#define CH_SIZE_X 25
#define CH_SIZE_Y 15
#define CH_SIZE_Z 25

/* WW...water works */
#define WW_SIZE_X 25
#define WW_SIZE_Y 25
#define WW_SIZE_Z 50

/* TR...tree */
#define TR_SIZE_X 10
#define TR_SIZE_Y 10
#define TR_SIZE_Z 15

/* BS...bush */
#define BS_SIZE_X 5
#define BS_SIZE_Y 5
#define BS_SIZE_Z 10
 
/* landscape size */
#define MAX 128

/* parser error messages */
#define OBJ_X_OVERFLOW  0
#define OBJ_X_UNDERFLOW 1
#define OBJ_Y_OVERFLOW  2
#define OBJ_Y_UNDERFLOW 3

/* error exit status code */
#define ERR_X_UNDERFLOW 2
#define ERR_X_OVERFLOW  3
#define ERR_Y_UNDERFLOW 4
#define ERR_Y_OVERFLOW  5

/* debug file */
#define DEBUG_FILE "debug.map"

/* set input landscape file name */
void SetLandscapeFileName(char *lndFileName);

/* set input texture container description file name */
void SetContainerFileName(char *cntFileName);

/* get parser error info */
void GetParserError(int objIndex, int message);

/* test object for coordinates overlay */
void TestObjectPlacement(void);

/* test all texture file */
void TestTextureFiles(void);

/* parse container file for objects, textures, placement */
/* and build internal scene database                     */
void ParseContainerFile(void);
                                                
/* main texture function, which map */
/* all objects to scene             */
void Map3DTexture(void);

#endif

/* End of file: object.h */
