/***************************************************************************/
/*                                                                         */
/*                m3dt - 3D texture file mapper v0.23                      */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*       parser.c - module to m3dt (contain auxiliary functions)           */
/*                                                                         */
/***************************************************************************/

/* include header files */

/* standard system header files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* user defined header files */
#include "parser.h"
#include "mpengine.h"

/* include header files specific for unix only */
#if defined(UNIX)
#include <unistd.h>
#include <sys/utsname.h>
#endif


/* global variables */

/* command line args for help */
char *strHelp = "-help";
char *strLongHelp = "--help";

/* function for putting message on terminal */
void PutMessage(int controlPoint){

	switch (controlPoint){

		case M3DT_START: /* say, that we are starting */
				 printf("%s", MSG_START);
				 break;
				 
		case M3DT_END:   /* say, that we are exiting */
				 printf("%s", MSG_END);
				 break;
				 
		case M3DT_USAGE: /* say, how to use this program */ 
				 printf("%s", MSG_USAGE);
				 printf("\n");
				 exit(OK);
				 break;
				 
		case M3DT_FILE_ERR: /* say, that we have problems with */
		                 /* opening  the input files           */ 
				 printf("%s", MSG_FILE_ERR);
				 break;
				 
		case M3DT_LFILE_ERR: /* say, that we have problems with */
		                 /* opening  the input landscape file   */ 
				 printf("%s", MSG_LFILE_ERR);
				 break;
				 
		case M3DT_CFILE_ERR: /* say, that we have problems with */
		                 /* opening  the input scene description*/ 
		                 /* file                                */ 
				 printf("%s", MSG_CFILE_ERR);
				 break;
				 
		case M3DT_TFILE_ERR: /* say, that we have problems with */
		                 /* opening  the input texture file     */ 
				 printf("%s", MSG_TFILE_ERR);
				 break;
				 
		default:         /* say, that we have a critical error */ 
				 printf("%s", MSG_ERR);
	}
}

/* function for setting up unix environment */
void InitEnvironment(void){

#if defined(UNIX)

	/* get miscellaneous unix environment infos */
	/* that is process PID, PPID, user UID, GID */
	printf("\n%s%d%s%d%c", "process id (", getpid(), "), process parent id (", getppid(), ')');
	printf("\n%s%d%s%d%c", "user uid (", getuid(), "), user gid (", getgid(), ')');

#endif
}

/* function for putting help infos on terminal */
void GetHelp(void){

	printf("%s", MSG_USAGE);
	printf("\n%s", "[landscape file] is a file, which contain landscape data");
	printf("\n%s", "[texture container file] is a file with all objects descriptions\n");
	exit(OK);
}

/* function for parsing command line arguments */
void ParseArg(int index, char *arg){

	/* test, if user want a help message */
	if (!strcmp(arg,strHelp) || !strcmp(arg,strLongHelp)){
		GetHelp();
		exit(OK);
	}

	switch (index){
		/* parse for input landscape file */
		case LANDSCAPE_FILE: SetLandscapeFileName(arg);
				     break;
				     
		/* parse for input texture container file */						   
		case CONTAINER_FILE: SetContainerFileName(arg);
				     break;

		default: return;
	}
}

/* End of file: contrib.c */
